import unittest
import os
from ..daltools import sirifc
from ..core import MolFrag

class TestException(unittest.TestCase):

    def setUp(self):
        self.thisdir = os.path.dirname(__file__)
        self.tmpdir = os.path.join(self.thisdir, 'h2o_sym', 'tmp')

    def test_raise_exception_if_symmetry(self):
        nsym = sirifc.sirifc(name=os.path.join(self.tmpdir, "SIRIFC")).nsym
        self.assertRaises(AssertionError, MolFrag, self.tmpdir)
        
    





from .common import LoPropTestCase
import os 
import sys
import numpy as np
from ..core import MolFrag
from ..daltools.util import full

import re
thisdir  = os.path.dirname(__file__)
case = "h2o"
tmpdir=os.path.join(thisdir, case, 'tmp')
exec('from . import %s_data as ref'%case)

from ..core import penalty_function, xtang, pairs


class NewTest(LoPropTestCase):

    def setUp(self):
        self.m = MolFrag(tmpdir, freqs=(0, ), pf=penalty_function(2.0/xtang**2))
        self.maxDiff = None

    def tearDown(self):
        pass


    def test_nuclear_charge(self):
        Z = self.m.Z
        self.assert_allclose(Z, ref.Z)

    def test_coordinates_au(self):
        R = self.m.R
        self.assert_allclose(R, ref.R)

    def test_default_gauge(self):
        self.assert_allclose(self.m.Rc, ref.Rc)

    def test_defined_gauge(self):
        m = MolFrag(tmpdir, gc=[1,2,3])
        self.assert_allclose(m.Rc, [1,2,3])

    def test_total_charge(self):
        Qtot = self.m.Qab.sum()
        self.assertAlmostEqual(Qtot, ref.Qtot)

    def test_charge(self):
        Qaa = self.m.Qa
        self.assert_allclose(ref.Q, Qaa)

    def test_total_dipole(self):
        self.assert_allclose(self.m.Dtot, ref.Dtot)

    def test_dipole_allbonds(self):
        D = full.matrix(ref.D.shape)
        Dab = self.m.Dab
        for ab, a, b in pairs(self.m.noa):
            D[:, ab] += Dab[:, a, b ] 
            if a != b: D[:, ab] += Dab[:, b, a] 
        self.assert_allclose(D, ref.D)

    def test_dipole_allbonds_sym(self):
        Dsym = self.m.Dsym
        self.assert_allclose(Dsym, ref.D)

    def test_dipole_nobonds(self):
        Daa = self.m.Dab.sum(axis=2).view(full.matrix)
        self.assert_allclose(Daa, ref.Daa)

    def test_quadrupole_total(self):
        QUc = self.m.QUc
        self.assert_allclose(QUc, ref.QUc)
    
    def test_nuclear_quadrupole(self):
        QUN = self.m.QUN
        self.assert_allclose(QUN, ref.QUN)

    def test_quadrupole_allbonds(self):
        QU = full.matrix(ref.QU.shape)
        QUab = self.m.QUab
        for ab, a, b in pairs(self.m.noa):
            QU[:, ab] += QUab[:, a, b ] 
            if a != b: QU[:, ab] += QUab[:, b, a] 
        self.assert_allclose(QU, ref.QU)

    def test_quadrupole_allbonds_sym(self):
        QUsym = self.m.QUsym
        self.assert_allclose(QUsym, ref.QU)

    def test_quadrupole_nobonds(self):
        self.assert_allclose(self.m.QUa, ref.QUaa)

    def test_Fab(self):
        Fab = self.m.Fab
        self.assert_allclose(Fab, ref.Fab)

    def test_molcas_shift(self):
        Fab = self.m.Fab
        Lab = Fab + self.m.sf(Fab)
        self.assert_allclose(Lab, ref.Lab)

    def test_total_charge_shift(self):
        dQ = self.m.dQa[0].sum(axis=0).view(full.matrix)
        dQref = [0., 0., 0.]
        self.assert_allclose(dQref, dQ)

    def test_atomic_charge_shift(self):
        dQa = self.m.dQa[0]
        dQaref = (ref.dQa[:, 1::2] - ref.dQa[:, 2::2])/(2*ref.ff)

        self.assert_allclose(dQa, dQaref, atol=.006)

    def test_lagrangian(self):
    # values per "perturbation" as in atomic_charge_shift below
        la = self.m.la[0]
        laref = (ref.la[:,0:6:2] - ref.la[:,1:6:2])/(2*ref.ff)
    # The sign difference is because mocas sets up rhs with opposite sign
        self.assert_allclose(-laref, la, atol=100)

    def test_bond_charge_shift(self):
        dQab = self.m.dQab[0]
        noa = self.m.noa


        dQabref = (ref.dQab[:, 1:7:2] - ref.dQab[:, 2:7:2])/(2*ref.ff)
        dQabcmp = full.matrix((3, 3))
        ab = 0
        for a in range(noa):
            for b in range(a):
                dQabcmp[ab, :] = dQab[a, b, :]
                ab += 1
    # The sign difference is because mocas sets up rhs with opposite sign
        self.assert_allclose(-dQabref, dQabcmp, atol=0.006)

    def test_bond_charge_shift_sum(self):
        dQa  = self.m.dQab[0].sum(axis=1).view(full.matrix)
        dQaref = self.m.dQa[0]
        self.assert_allclose(dQa, dQaref)


    def test_polarizability_total(self):

        Am = self.m.Am[0]

        self.assert_allclose(Am, ref.Am, 0.015)

    def notest_dynamic_polarizability_total(self):

        Amw = self.m.Am[1]

        self.assert_allclose(Amw, ref.Amw, 0.015)
            
    def test_polarizability_allbonds_molcas_internal(self):

        O = ref.O
        H1O = ref.H1O
        H1 = ref.H1
        H2O = ref.H2O
        H2H1 = ref.H2H1
        H2 = ref.H2
        rMP = ref.rMP

        RO, RH1, RH2 = self.m.R
        ROx, ROy, ROz = RO
        RH1x, RH1y, RH1z = RH1
        RH2x, RH2y, RH2z = RH2
        

        ihff = 1/(2*ref.ff)

        q, x, y, z = range(4)
        dx1, dx2, dy1, dy2, dz1, dz2 = 1, 2, 3, 4, 5, 6
        o, h1o, h1, h2o, h2h1, h2 = range(6)

        Oxx = ihff*(rMP[x, dx1, o] - rMP[x, dx2, o])
        Oyx = ihff*(rMP[y, dx1, o] - rMP[y, dx2, o]
            +       rMP[x, dy1, o] - rMP[x, dy2, o])/2
        Oyy = ihff*(rMP[y, dy1, o] - rMP[y, dy2, o])
        Ozx = ihff*(rMP[z, dx1, o] - rMP[z, dx2, o]
            +       rMP[x, dz1, o] - rMP[x, dz2, o])/2
        Ozy = ihff*(rMP[z, dy1, o] - rMP[z, dy2, o]
            +       rMP[y, dz1, o] - rMP[y, dz2, o])/2
        Ozz = ihff*(rMP[z, dz1, o] - rMP[z, dz2, o])
        H1Oxx = ihff*(rMP[x, dx1, h1o] - rMP[x, dx2, h1o] \
              - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1x-ROx))
        H1Oyx = ihff*(
             (rMP[y, dx1, h1o] - rMP[y, dx2, h1o] 
            + rMP[x, dy1, h1o] - rMP[x, dy2, h1o])/2
           - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1y-ROy)
    #      - (rMP[0, dy1, h1o] - rMP[0, dy2, h1o])*(RH1x-ROx) THIS IS REALLY... A BUG?
           )
        H1Oyy = ihff*(rMP[y, dy1, h1o] - rMP[y, dy2, h1o] - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH1y-ROy))
        H1Ozx = ihff*(
            (rMP[z, dx1, h1o] - rMP[z, dx2, h1o]
           + rMP[x, dz1, h1o] - rMP[x, dz2, h1o])/2
          - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1z-ROz)
    #             - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1x-ROx) #THIS IS REALLY... A BUG?
                )
        H1Ozy = ihff*(
            (rMP[z, dy1, h1o] - rMP[z, dy2, h1o]
           + rMP[y, dz1, h1o] - rMP[y, dz2, h1o])/2
          - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH1z-ROz)
    #     - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1y-ROy) THIS IS REALLY... A BUG?
            )
        H1Ozz = ihff*(rMP[z, dz1, h1o] - rMP[z, dz2, h1o] - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1z-ROz))
        H1xx = ihff*(rMP[x, dx1, h1] - rMP[x, dx2, h1])
        H1yx = (ihff*(rMP[y, dx1, h1] - rMP[y, dx2, h1])
             +  ihff*(rMP[x, dy1, h1] - rMP[x, dy2, h1]))/2
        H1yy = ihff*(rMP[y, dy1, h1] - rMP[y, dy2, h1])
        H1zx = (ihff*(rMP[z, dx1, h1] - rMP[z, dx2, h1])
             +  ihff*(rMP[x, dz1, h1] - rMP[x, dz2, h1]))/2
        H1zy = (ihff*(rMP[z, dy1, h1] - rMP[z, dy2, h1])
             +  ihff*(rMP[y, dz1, h1] - rMP[y, dz2, h1]))/2
        H1zz = ihff*(rMP[z, dz1, h1] - rMP[z, dz2, h1])
        H2Oxx = ihff*(rMP[x, dx1, h2o] - rMP[x, dx2, h2o] - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2x-ROx))
        H2Oyx = ihff*(
            (rMP[y, dx1, h2o] - rMP[y, dx2, h2o] 
           + rMP[x, dy1, h2o] - rMP[x, dy2, h2o])/2
           - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2y-ROy)
    #      - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH2x-ROx) THIS IS REALLY... A BUG?
           )
        H2Oyy = ihff*(rMP[y, dy1, h2o] - rMP[y, dy2, h2o] - (rMP[q, dy1, h2o] - rMP[q, dy2, h2o])*(RH2y-ROy))
        H2Ozx = ihff*(
            (rMP[z, dx1, h2o] - rMP[z, dx2, h2o]
           + rMP[x, dz1, h2o] - rMP[x, dz2, h2o])/2
          - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2z-ROz)
    #             - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH2x-ROx) #THIS IS REALLY... A BUG?
                )
        H2Ozy = ihff*(
            (rMP[z, dy1, h2o] - rMP[z, dy2, h2o]
           + rMP[y, dz1, h2o] - rMP[y, dz2, h2o])/2
          - (rMP[q, dy1, h2o] - rMP[q, dy2, h2o])*(RH2z-ROz)
    #     - (rMP[q, dz1, h2o] - rMP[q, dz2, h2o])*(RH2y-ROy) THIS IS REALLY... A BUG?
            )
        H2Ozz = ihff*(rMP[z, dz1, h2o] - rMP[z, dz2, h2o] - (rMP[q, dz1, h2o] - rMP[q, dz2, h2o])*(RH2z-ROz))
        H2H1xx = ihff*(rMP[x, dx1, h2h1] - rMP[x, dx2, h2h1] - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH2x-RH1x))
        H2H1yx = ihff*(
            (rMP[y, dx1, h2h1] - rMP[y, dx2, h2h1] 
           + rMP[x, dy1, h2h1] - rMP[x, dy2, h2h1])/2
           - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH1y-ROy)
    #      - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH1x-ROx) THIS IS REALLY... A BUG?
           )
        H2H1yy = ihff*(rMP[y, dy1, h2h1] - rMP[y, dy2, h2h1] - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH2y-RH1y))
        H2H1zx = ihff*(
            (rMP[z, dx1, h2h1] - rMP[z, dx2, h2h1]
           + rMP[x, dz1, h2h1] - rMP[x, dz2, h2h1])/2
          - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH1z-ROz)
    #     - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH1x-ROx) #THIS IS REALLY... A BUG?
                )
        H2H1zy = ihff*(
            (rMP[z, dy1, h2h1] - rMP[z, dy2, h2h1]
           + rMP[y, dz1, h2h1] - rMP[y, dz2, h2h1])/2
          - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH1z-ROz)
    #     - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH1y-RO[1]) THIS IS REALLY... A BUG?
            )
        H2H1zz = ihff*(rMP[z, dz1, h2h1] - rMP[z, dz2, h2h1] - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH2z-RH1z))
        H2xx = ihff*(rMP[x, dx1, h2] - rMP[x, dx2, h2])
        H2yx = (ihff*(rMP[y, dx1, h2] - rMP[y, dx2, h2])
             +  ihff*(rMP[x, dy1, h2] - rMP[x, dy2, h2]))/2
        H2yy = ihff*(rMP[y, dy1, h2] - rMP[y, dy2, h2])
        H2zx = (ihff*(rMP[z, dx1, h2] - rMP[z, dx2, h2])
             +  ihff*(rMP[x, dz1, h2] - rMP[x, dz2, h2]))/2
        H2zy = (ihff*(rMP[z, dy1, h2] - rMP[z, dy2, h2])
             +  ihff*(rMP[y, dz1, h2] - rMP[y, dz2, h2]))/2
        H2zz = ihff*(rMP[z, dz1, h2] - rMP[z, dz2, h2])

        comp = ("XX", "yx", "yy", "zx", "zy", "zz")
        bond = ("O", "H1O", "H1", "H2O", "H2H1", "H2")

        self.assert_allclose(O[0], Oxx, text="Oxx")
        self.assert_allclose(O[1], Oyx, text="Oyx")
        self.assert_allclose(O[2], Oyy, text="Oyy")
        self.assert_allclose(O[3], Ozx, text="Ozx")
        self.assert_allclose(O[4], Ozy, text="Ozy")
        self.assert_allclose(O[5], Ozz, text="Ozz")
        self.assert_allclose(H1O[0], H1Oxx, text="H1Oxx")
        self.assert_allclose(H1O[1], H1Oyx, text="H1Oyx")
        self.assert_allclose(H1O[2], H1Oyy, text="H1Oyy")
        self.assert_allclose(H1O[3], H1Ozx, text="H1Ozx")
        self.assert_allclose(H1O[4], H1Ozy, text="H1Ozy")
        self.assert_allclose(H1O[5], H1Ozz, text="H1Ozz")
        self.assert_allclose(H1[0], H1xx, text="H1xx")
        self.assert_allclose(H1[1], H1yx, text="H1yx")
        self.assert_allclose(H1[2], H1yy, text="H1yy")
        self.assert_allclose(H1[3], H1zx, text="H1zx")
        self.assert_allclose(H1[4], H1zy, text="H1zy")
        self.assert_allclose(H1[5], H1zz, text="H1zz")
        self.assert_allclose(H2O[0], H2Oxx, text="H2Oxx")
        self.assert_allclose(H2O[1], H2Oyx, text="H2Oyx")
        self.assert_allclose(H2O[2], H2Oyy, text="H2Oyy")
        self.assert_allclose(H2O[3], H2Ozx, text="H2Ozx")
        self.assert_allclose(H2O[4], H2Ozy, text="H2Ozy")
        self.assert_allclose(H2O[5], H2Ozz, text="H2Ozz")
        self.assert_allclose(H2H1[0], H2H1xx, text="H2H1xx")
        self.assert_allclose(H2H1[1], H2H1yx, text="H2H1yx")
        self.assert_allclose(H2H1[2], H2H1yy, text="H2H1yy")
        self.assert_allclose(H2H1[3], H2H1zx, text="H2H1zx")
        self.assert_allclose(H2H1[4], H2H1zy, text="H2H1zy")
        self.assert_allclose(H2H1[5], H2H1zz, text="H2H1zz")
        self.assert_allclose(H2[0], H2xx, text="H2xx")
        self.assert_allclose(H2[1], H2yx, text="H2yx")
        self.assert_allclose(H2[2], H2yy, text="H2yy")
        self.assert_allclose(H2[3], H2zx, text="H2zx")
        self.assert_allclose(H2[4], H2zy, text="H2zy")
        self.assert_allclose(H2[5], H2zz, text="H2zz")

    def test_altint(self):
        R = self.m.R
        rMP = ref.rMP
        diff = [(1, 2), (3, 4), (5, 6)]
        atoms = (0, 2, 5) 
        bonds = (1, 3, 4)
        ablab = ("O", "H1O", "H1", "H2O", "H2H1", "H2")
        ijlab = ("xx", "yx", "yy", "zx", "zy", "zz")

        pol = np.zeros((6, self.m.noa*(self.m.noa+1)//2))
        for ab, a, b in pairs(self.m.noa):
            for ij, i, j in pairs(3):
                #from pdb import set_trace; set_trace()
                i1, i2 = diff[i]
                j1, j2 = diff[j]
                pol[ij, ab] += (rMP[i+1, j1, ab] - rMP[i+1, j2, ab]
                            +   rMP[j+1, i1, ab] - rMP[j+1, i2, ab])/(4*ref.ff)
                if ab in bonds:
                    pol[ij, ab] -= (R[a][i]-R[b][i])*(rMP[0, j1, ab] - rMP[0, j2, ab])/(2*ref.ff)
                self.assert_allclose(ref.Aab[ij, ab], pol[ij, ab], text="%s%s"%(ablab[ab], ijlab[ij]))

    def test_polarizability_allbonds_atoms(self):

        Aab = self.m.Aab[0] #+ m.dAab
        noa = self.m.noa

        Acmp=full.matrix(ref.Aab.shape)
        
        ab = 0
        for a in range(noa):
            for b in range(a):
                Acmp[:, ab] = (Aab[:, :, a, b] + Aab[:, :, b, a]).pack()
                ab += 1
            Acmp[:, ab] = Aab[:, :, a, a].pack()
            ab += 1
        # atoms
        self.assert_allclose(ref.Aab[:, 0], Acmp[:, 0], atol=.005)
        self.assert_allclose(ref.Aab[:, 2], Acmp[:, 2], atol=.005)
        self.assert_allclose(ref.Aab[:, 5], Acmp[:, 5], atol=.005)

    def test_polarizability_allbonds_bonds(self):

        Aab = self.m.Aab[0] + self.m.dAab[0]/2
        noa = self.m.noa

        Acmp=full.matrix(ref.Aab.shape)
        
        ab = 0
        for a in range(noa):
            for b in range(a):
                Acmp[:, ab] = (Aab[:, :, a, b] + Aab[:, :, b, a]).pack()
                ab += 1
            Acmp[:, ab] = Aab[:, :, a, a].pack()
            ab += 1
        # atoms
        self.assert_allclose(ref.Aab[:, 1], Acmp[:, 1], atol=.150)
        self.assert_allclose(ref.Aab[:, 3], Acmp[:, 3], atol=.150)
        self.assert_allclose(ref.Aab[:, 4], Acmp[:, 4], atol=.005)
        

    def test_polarizability_nobonds(self):

        Aab = self.m.Aab[0] + self.m.dAab[0]/2
        noa = self.m.noa

        Acmp = full.matrix((6, noa ))
        Aa = Aab.sum(axis=3).view(full.matrix)
        
        ab = 0
        for a in range(noa):
            Acmp[:, a,] = Aa[:, :, a].pack()

        # atoms
        self.assert_allclose(Acmp, ref.Aa, atol=0.07)

    def test_potfile_PAn0(self):
        PAn0 = self.m.output_potential_file(maxl=-1, pol=0, hyper=0)
        self.assert_str(PAn0, ref.PAn0)

    def test_potfile_PA00(self):
        PA00 = self.m.output_potential_file(maxl=0, pol=0, hyper=0)
        self.assert_str(PA00, ref.PA00)

    def test_potfile_PA10(self):
        PA10 = self.m.output_potential_file(maxl=1, pol=0, hyper=0)
        self.assert_str(PA10, ref.PA10)

    def test_potfile_PA20(self):
        PA20 = self.m.output_potential_file(maxl=2, pol=0, hyper=0)
        self.assert_str(PA20, ref.PA20)

    def test_potfile_PA21(self):
        PA21 = self.m.output_potential_file(maxl=2, pol=1, hyper=0)
        self.assert_str(PA21, ref.PA21)

    def test_potfile_PA22(self):
        PA22 = self.m.output_potential_file(maxl=2, pol=2, hyper=0)
        self.assert_str(PA22, ref.PA22)

    def test_potfile_PAn0b(self):
        PAn0b = self.m.output_potential_file(maxl=-1, pol=0, hyper=0, bond_centers=True)
        self.assert_str(PAn0b, ref.PAn0b)

    def test_potfile_PA00b(self):
        PA00b = self.m.output_potential_file(maxl=0, pol=0, hyper=0, bond_centers=True)
        self.assert_str(PA00b, ref.PA00b)

    def test_potfile_PA10b(self):
        PA10b = self.m.output_potential_file(maxl=1, pol=0, hyper=0, bond_centers=True)
        self.assert_str(PA10b, ref.PA10b)

    def test_potfile_PA20b(self):
        with self.assertRaises(NotImplementedError):
            PA20b = self.m.output_potential_file(maxl=2, pol=0, hyper=0, bond_centers=True)

    def test_potfile_PA01b(self):
        PA01b = self.m.output_potential_file(maxl=0, pol=1, hyper=0, bond_centers=True)
        self.assert_str(PA01b, ref.PA01b)

    def test_potfile_PA02(self):
        this = self.m.output_potential_file(maxl=0, pol=2, hyper=0)
        self.assert_str(this, ref.PA02)

    def test_potfile_PA02b(self):
        this = self.m.output_potential_file(maxl=0, pol=2, hyper=0, bond_centers=True)
        self.assert_str(this, ref.PA02b)

    def test_outfile_PAn0_atom_domain(self):
        self.m.max_l = -1
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_n0_1)


    def test_outfile_PA00_atom_domain(self):
        self.m.max_l = 0
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_00_1)

    def test_outfile_PA10_atom_domain(self):
        self.m.max_l = 1
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_10_1)

    def test_outfile_PA20_atom_domain(self):
        self.m.max_l = 2
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_20_1)

    def test_outfile_PA01_atom_domain(self):
        self.m.max_l = 0
        self.m.pol = 1
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_01_1)

    def test_outfile_PA02_atom_domain(self):
        self.m.max_l = 0
        self.m.pol = 2
        self.assert_str(self.m.print_atom_domain(0), ref.OUTPUT_02_1)

    def test_outfile_PAn0_by_atom(self):
        self.m.max_l = -1
        self.m.output_by_atom(fmt="%12.5f")
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_n0)

    def test_outfile_PAn0_by_atom_Angstrom(self):
        self.m.max_l = -1
        self.m.output_by_atom(fmt="%12.5f", angstrom=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_n0_ANGSTROM)

    def test_outfile_PA00_by_atom(self):
        self.m.output_by_atom(fmt="%12.5f", max_l=0)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_00)

    def test_outfile_PA10_by_atom(self):
        self.m.max_l = 1
        self.m.output_by_atom(fmt="%12.5f", max_l=1)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_10)

    def test_outfile_PA20_by_atom(self):
        self.m.max_l = 2
        self.m.output_by_atom(fmt="%12.5f", max_l=2)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_20)

    def test_outfile_PA01_by_atom(self):
        self.m.max_l = 0
        self.m.output_by_atom(fmt="%12.5f", max_l=0, pol=1)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_ATOM_01)

    def test_outfile_PAn0_by_bond(self):
        self.m.max_l = -1
        self.m.output_by_atom(fmt="%12.5f", max_l=-1, bond_centers=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_BOND_n0)

    def test_outfile_PA00_by_bond(self):
        self.m.max_l = 0
        self.m.output_by_atom(fmt="%12.5f", max_l=0, bond_centers=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_BOND_00)

    def test_outfile_PA10_by_bond(self):
        self.m.max_l = 1
        self.m.output_by_atom(fmt="%12.5f", max_l=1, bond_centers=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_BOND_10)

    def test_outfile_PA10_by_bond(self):
        with self.assertRaises(NotImplementedError):
            self.m.output_by_atom(fmt="%12.5f", max_l=2, bond_centers=True)

    def test_outfile_PAn1_by_bond(self):
        self.m.max_l = -1
        self.m.output_by_atom(fmt="%12.5f", max_l=-1, pol=1, bond_centers=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_BOND_n1)

    def test_outfile_PAn2_by_bond(self):
        self.m.max_l = -1
        self.m.output_by_atom(fmt="%12.5f", max_l=-1, pol=2, bond_centers=True)
        print_output = sys.stdout.getvalue().strip()
        self.assert_str(print_output, ref.OUTPUT_BY_BOND_n2)

if __name__ == "__main__":
    pass



import unittest
try:
    import mock
except ImportError: # pragma: no cover
    from unittest import mock
import numpy
import math
from ..blocked import *
from ..full import init

class TestBlocked(unittest.TestCase):

    def setUp(self):
        self.bdm = BlockDiagonalMatrix((2, 1), (2, 1))

    def tearDown(self):
        pass

    def assert_allclose(self, this, ref):
        for t, r in zip(this, ref):
            numpy.testing.assert_allclose(t, r)

    def test_rowdim(self):
        self.assertTupleEqual(self.bdm.nrow, (2, 1))

    def test_coldim(self):
        self.assertTupleEqual(self.bdm.ncol, (2, 1))

    def test_row_offset(self):
        self.assertTupleEqual(self.bdm.irow, (0, 2))

    def test_col_offset(self):
        self.assertTupleEqual(self.bdm.icol, (0, 2))

    def test_incompatible_dim_raises(self):
        self.assertRaises(AssertionError, BlockDiagonalMatrix, (1, 2), (3,))

    def test_str(self):
        self.assertEqual(str(self.bdm), """
Block 1

 (2, 2) 
              Column   1    Column   2

Block 2

 (1, 1) 
              Column   1
""")


    def test_init(self):
        self.bdm[0][:, :] = [[1, 3], [2, 4]]
        self.bdm[1][:, :] = [[5]]
        bdm = BlockDiagonalMatrix.init([[1, 2], [3, 4]], [[5]])
        numpy.testing.assert_allclose(bdm.subblock[0], self.bdm.subblock[0])
        numpy.testing.assert_allclose(bdm.subblock[1], self.bdm.subblock[1])


    def test_init_from_array_dimension_error(self):
        with self.assertRaises(AssertionError):
            bdm = BlockDiagonalMatrix.init_from_array(
                [1, 2, 3], (2, 1), (2, 1)
                )

    def test_init_from_array(self):
        self.bdm[0][:, :] = [[0, 2], [1, 3]]
        self.bdm[1][:, :] = [[4]]
        bdm = BlockDiagonalMatrix.init_from_array(range(5), (2, 1), (2, 1))
        numpy.testing.assert_allclose(bdm.subblock[0], self.bdm.subblock[0])
        numpy.testing.assert_allclose(bdm.subblock[1], self.bdm.subblock[1])

    def test_blocked_ravel(self):
        bdm = BlockDiagonalMatrix.init_from_array(range(5), (2, 1), (2, 1))
        numpy.testing.assert_allclose(bdm.ravel(order='F'), range(5))
        

    def test_unblock(self):
        blocked = BlockDiagonalMatrix((2, 1), (2, 1))
        blocked[0][:, :] = ((1, 2), (3, 4))
        blocked[1][:, :] = ((5,),)
        unblocked = [[1, 2, 0], [3, 4, 0], [0, 0, 5]]
        numpy.testing.assert_allclose(blocked.unblock(), unblocked)

    def test_pack(self):
        blocked = BlockDiagonalMatrix((2, 1), (2, 1))
        blocked[0][:, :] = ((1, 2), (2, 1))
        blocked[1][:, :] = ((5,),)
        packed = [[1, 2, 1], [5]]
        self.assert_allclose(blocked.pack().subblock, packed)
        

    def test_unit(self):
        ref = [[[1, 0], [0, 1]], [[1]]]
        blocked_unit = unit((2, 1))
        self.assert_allclose(blocked_unit, ref)

    def test_mul(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(self.bdm*self.bdm, [[[2, 3], [6, 11]], [[4]]])

    def test_rmul(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(2*self.bdm, [[[0, 2], [4, 6]], [[4]]])

    def test_add(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(self.bdm + self.bdm, [[[0, 2], [4, 6]], [[4]]])

    def test_sub(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(self.bdm - self.bdm, [[[0, 0], [0, 0]], [[0]]])

    def test_neg(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(-self.bdm, [[[0, -1], [-2, -3]], [[-2]]])

    def test_div(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(self.bdm/self.bdm, [[[1, 0], [0, 1]], [[1]]])

    def test_scalar_div(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(self.bdm/2, [[[0, .5], [1, 1.5]], [[1]]])

    def test_rdiv(self):
        self.bdm[0][:, :] = [[0, 1], [2, 3]]
        self.bdm[1][:, :] = [[2]]
        self.assert_allclose(1/self.bdm, [[[-1.5, 0.5], [1, 0]], [[0.5]]])

    def test_sqrt(self):
        self.bdm[0][:, :] = [[4, 0], [0, 4]]
        self.bdm[1][:, :] = [[9]]
        self.assert_allclose(self.bdm.sqrt(), [[[2, 0], [0, 2]], [[3]]])

    def test_isqrt(self):
        self.bdm[0][:, :] = [[4, 0], [0, 4]]
        self.bdm[1][:, :] = [[9]]
        self.assert_allclose(self.bdm.invsqrt(), [[[.5, 0], [0, .5]], [[1./3]]])

    def test_eigvec(self):
        self.bdm[0][:, :] = [[4, 0], [0, 4]]
        self.bdm[1][:, :] = [[9]]
        u, v = self.bdm.eigvec()
        self.assert_allclose(u, [4, 9])
        self.assert_allclose(v, [[[1, 0], [0, 1]], [[1]]])

    def test_transpose(self):
        self.bdm[0][:, :] = [[1, 2], [3, 4]]
        bdm_T = self.bdm.T()
        self.assert_allclose(bdm_T[0], [[1, 3], [2, 4]])

    def test_get_columns(self):
        self.bdm[0][:, :] = [[1, 2], [3, 4]]
        subset = self.bdm.get_columns((1, 0))
        self.assert_allclose(subset[0], [1, 3])

    def test_func(self):
        self.bdm[0][:, :] = [[1, 0], [0, 16]]
        self.bdm[1][:, :] = [[25]]
        self.assert_allclose(self.bdm.func(math.sqrt), [[[1, 0], [0, 4]], [[5]]])

    @mock.patch.object(numpy.random, 'random')
    def test_blocked_random(self, mock_random):
        M = BlockDiagonalMatrix([2], [2]).random()
        self.assertTrue(mock_random.calls, 2)

    def test_tr(self):
        M = BlockDiagonalMatrix([2, 1], [2, 1])
        M.subblock[0][0, 0] = 3
        M.subblock[0][1, 1] = 2
        M.subblock[1][0, 0] = 1
        self.assertEqual(M.tr(), 6.0)

    def test_qr_Q(self):
        A = BlockDiagonalMatrix([3], [3])
        A.subblock[0] = init([[12, 6, -4], [-51, 167, 24], [4, -68, -41]])
        Q, R = A.qr()
        numpy.testing.assert_almost_equal(Q[0], 
            -init([[6./7, 3./7, -2./7], [-69./175, 158./175, 6./35], [-58./175, 6./175, -33./35]])
            )
    def test_qr_R(self):
        A = BlockDiagonalMatrix([3], [3])
        A.subblock[0] = init([[12, 6, -4], [-51, 167, 24], [4, -68, -41]])
        Q, R = A[0].qr()
        numpy.testing.assert_almost_equal(R, 
            -init([[14, 0, 0], [21, 175, 0], [-14, -70, 35]])
            )
        numpy.testing.assert_almost_equal(A[0], Q*R)

    def test_gram_schmidt(self):
        S = BlockDiagonalMatrix([2], [2])
        v = BlockDiagonalMatrix([2], [2])
        Delta = 0.1
        S.subblock[0] = init([[1.0, Delta], [Delta, 1.0]])
        v.subblock[0] = init([[1.0, 0.0], [0.0, 1.0]])
        u = v.GS(S)
        u_ref = init([[1.0, 0.0], [-Delta/math.sqrt(1-Delta**2), 1.0/math.sqrt(1-Delta**2)]])
        numpy.testing.assert_almost_equal(u[0], u_ref)
        
class BlockedTriangularTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_str(self):
        bt = triangular((2, 1))
        self.assertEqual(str(bt), """
Block 1

    0.00000000
    0.00000000    0.00000000

Block 2

    0.00000000
"""
        ) 

    def test_init(self):
        bt = triangular.init([[1., 2., 3.], [4.]])
        numpy.testing.assert_almost_equal(bt.subblock[0], [1., 2., 3.])
        numpy.testing.assert_almost_equal(bt.subblock[1], [4.])

    @mock.patch.object(numpy.random, 'random')
    def test_blocked_random(self, mock_random):
        M = triangular([3, 2]).random()
        self.assertTrue(mock_random.calls, 2)
        
    def test_add(self):
        bt = triangular.init([[1., 2., 3.], [4.]])
        bt = bt + bt
        numpy.testing.assert_almost_equal(bt.subblock[0], [2., 4., 6.])
        numpy.testing.assert_almost_equal(bt.subblock[1], [8.])

    def test_sub(self):
        bt = triangular.init([[1., 2., 3.], [4.]])
        bt = bt - bt
        numpy.testing.assert_almost_equal(bt.subblock[0], [0., 0., 0.])
        numpy.testing.assert_almost_equal(bt.subblock[1], [0.])

    def test_unpack(self):
        bt = triangular.init([[1., 2., 3.], [4.]])
        ubt = bt.unpack()
        numpy.testing.assert_almost_equal(ubt.subblock[0], [[1, 2], [2, 3]])
        numpy.testing.assert_almost_equal(ubt.subblock[1], [[4]])

    def test_unblock(self):
        bt = triangular.init([[1., 2., 3.], [4.]])
        ubl = bt.unblock()
        numpy.testing.assert_almost_equal(ubl, [1., 2., 3., 0., 0., 4.])


if __name__ == "__main__": # pragma: no cover
    unittest.main()



#!/usr/bin/env python
"""
Loprop model implementation (J. Chem. Phys. 121, 4494 (2004))
"""
import os
import sys
import math
import numpy
from functools import reduce
from .daltools import one, mol, dens, prop, lr, qr, sirifc
from .daltools.util import full, blocked, subblocked, timing

#full.matrix.fmt = "%14.6f"
xtang = 0.5291772108
angtx = 1.0/xtang

# Bragg-Slater radii () converted from Angstrom to Bohr
rbs = numpy.array([0, 
      0.25,                                     0.25, 
      1.45, 1.05, 0.85, 0.70, 0.65, 0.60, 0.50, 0.45,
      1.80, 1.50, 1.25, 1.10, 1.00, 1.00, 1.00, 1.00, 
      ])*angtx

bond_co = { ( 1, 1 ) : 1.2, 
            ( 1, 6 ) : 1.2, 
            ( 6, 1 ) : 1.2, 
            ( 1, 7 ) : 1.2, 
            ( 7, 1 ) : 1.2, 
            ( 1, 8 ) : 1.2, 
            ( 8, 1 ) : 1.2, 
            ( 1, 15) : 1.2, 
            ( 15, 1) : 1.2, 
            ( 6, 6 ) : 1.6, 
            ( 6, 7 ) : 1.6, 
            ( 7, 6 ) : 1.6, 
            ( 6, 8 ) : 1.6, 
            ( 8, 6 ) : 1.6, 
            ( 6, 15) : 1.6, 
            ( 15, 6) : 1.6, 
            ( 7, 7 ) : 1.6, 
            ( 7, 8 ) : 1.6, 
            ( 8, 7 ) : 1.6, 
            ( 7, 15) : 1.6, 
            ( 15, 7) : 1.6, 
            ( 8, 8 ) : 1.6, 
            ( 8, 15) : 1.6, 
            ( 15, 8) : 1.6, 
            ( 15, 15 ) : 1.6,
            }
#permute dict items in keypairs
#for key1, key2 in bond_co.keys():
    #bond_co[ ( key2, key1 ) ] = bond_co[ (key1, key2) ]

def symmetrize_first_beta( beta ):
# naive solution, transforms matrix B[ (x,y,z) ][ (xx, xy, xz, yy, yz, zz) ] into array
# Symmtrized UT array    B[ (xxx, xxy, xxz, xyy, xyz, xzz, yyy, yyz, yzz, zzz) ]

    new = full.matrix( 10 )

    new[0] = beta[0,0]
    new[1] = (beta[0,1] + beta[1,0] ) /2
    new[2] = (beta[0,2] + beta[2,0] ) /2
    new[3] = (beta[0,3] + beta[1,1] ) /2
    new[4] = (beta[0,4] + beta[1,2] + beta[2,1] ) /3
    new[5] = (beta[0,5] + beta[2,2] ) /2
    new[6] = beta[1,3] 
    new[7] = (beta[1,4] + beta[2,3] ) /2
    new[8] = (beta[1,5] + beta[2,4] ) /2
    new[9] = beta[2,5]

    return new

def penalty_function(alpha=2):
    """Returns function object """
    def pf(Za, Ra, Zb, Rb):
        """Inverse half of penalty function defined in Gagliardi"""

        from math import exp
        ra = rbs[int(round(Za))]
        rb = rbs[int(round(Zb))]

        xa, ya, za = Ra
        xb, yb, zb = Rb
        rab2 = (xa - xb)**2 + (ya - yb)**2 + (za - zb)**2

        f = 0.5*exp(-alpha*(rab2/(ra+rb)**2))
        return f
    return pf

def pairs(n):
    """Generate index pairs for triangular packed matrices up to n """
    ij = 0
    for i in range(n):
        for j in range(i+1):
            yield (ij, i, j)
            ij += 1

def shift_function(*args):
    """Return value twice max value of F"""
    F, = args
    return 2*numpy.max(numpy.abs(F))

def header(string):
    """Pretty print header"""
    border = '-'*len(string)
    print("\n%s\n%s\n%s" % (border, string, border))

def output_beta(beta, dip=None, fmt="%12.6f"):
    """Repeated output format for b(x; yz)"""
    print("Hyperpolarizability")
    print("beta(:, xx xy xz yy yz zz)")
    print("--------------------------")
    print("beta(x, *) " + (6*fmt) % tuple(beta[0,:]))
    print("beta(y, *) " + (6*fmt) % tuple(beta[1,:]))
    print("beta(z, *) " + (6*fmt) % tuple(beta[2,:]))
    betakk = beta[:,0] + beta[:, 3] + beta[:, 5]
    print("beta(:, kk)" + (3*fmt) % tuple(betakk))
    if dip is not  None:
        betapar = 0.2*(betakk & dip)/dip.norm2()
        print("beta//dip  " + (fmt) % betapar)

class MolFrag:
    """An instance of the MolFrag class is created and populated with
    data from a Dalton runtime scratch directory"""

    def __init__(
        self, tmpdir, max_l=0, pol=0, freqs=None, pf=penalty_function, sf=shift_function, gc=None,
        damping=False, real_pol=False, imag_pol=False, debug=False
        ):
        """Constructur of MolFrac class objects
        input: tmpdir, scratch directory of Dalton calculation
        """
        self.max_l = max_l
        self.pol = pol
        self.tmpdir = tmpdir
        if freqs is None:
            self.freqs = (0,)
            self.nfreqs = 1
        else:
            self.freqs = freqs
            self.nfreqs = len(freqs)
        self.rfreqs = range(self.nfreqs)
        self.damping = damping
        self._real_pol = real_pol
        self._imag_pol = imag_pol

        self.pf = pf
        self.sf = sf
        self.gc = gc
        self.debug = debug
        #
        # Dalton files
        #
        self.aooneint = os.path.join(tmpdir,'AOONEINT')
        self.dalton_bas = os.path.join(tmpdir,'DALTON.BAS')
        self.sirifc = os.path.join(tmpdir,'SIRIFC')
        assert sirifc.sirifc(name=self.sirifc).nsym == 1

        self._T = None
        self._D = None
        self._Dk = None
        self._D2k = None
        self.get_basis_info()
        self.get_isordk()
        self._x = None

        self._Rc = None
        self._Qab = None
        self._Da = None
        self._Dab = None
        self._Dsym = None
        self._QUab = None
        self._QUa = None
        self._QUsym = None
        self._QUN = None
        self._QUc = None
        self._dQa = None
        self._d2Qa = None
        self._dQab = None
        self._d2Qab = None
        self._Fab = None
        self._la = None
        self._l2a = None
        self._Aab = None
        self._Bab = None
        self._Am = None
        self._Bm = None
        self._dAab = None
        self._dBab = None
        #if maxl >= 0: self.charge()
        #if maxl >= 1: self.dipole()
        #if maxl >= 2: self.quadrupole()
        #if pol: self.pol()

    @property
    def real_pol(self):
        return self._real_pol

    @property
    def imag_pol(self):
        return self._imag_pol

    def set_real_pol(self):
        self._real_pol = True
        self._imag_pol = False

    def set_imag_pol(self):
        self._real_pol = False
        self._imag_pol = True


    def get_basis_info(self):
        """ Obtain basis set info from DALTON.BAS """
        molecule = mol.readin(self.dalton_bas)
        self.cpa = mol.contracted_per_atom(molecule)
        self.cpa_l = mol.contracted_per_atom_l(molecule)
        self.opa = mol.occupied_per_atom(molecule)
        self.noa = len(self.opa)
#
# Total number of basis functions and occpied orbitals
#
        self.nbf = sum(self.cpa)
        self.noc = 0
        for o in self.opa:
            self.noc += len(o)

        if self.debug:#pragma: no cover
            print("Orbitals/atom", self.cpa, "\nTotal", self.nbf)
            print("Occupied/atom", self.opa, "\nTotal", self.noc)

    def S(self):
        """
        Get overlap, nuclear charges and coordinates from AOONEINT
        """
        S = one.read("OVERLAP", self.aooneint)
        return  S.unpack().unblock()
        

    def get_isordk(self):
        """
        Get overlap, nuclear charges and coordinates from AOONEINT
        """
        #
        # Data from the ISORDK section in AOONEINT
        #
        isordk = one.readisordk(filename=self.aooneint)
        #
        # Number of nuclei
        #
        N = isordk["nucdep"]
        #
        # MXCENT , Fix dimension defined in nuclei.h
        #
        mxcent = len(isordk["chrn"]) 
        #
        # Nuclear charges
        #
        self.Z = full.matrix((N,))
        self.Z[:] = isordk["chrn"][:N]
        #
        # Nuclear coordinates
        #
        R = full.matrix((mxcent*3,))
        R[:] = isordk["cooo"][:]
        self.R = R.reshape((mxcent, 3), order='F')[:N, :]
       #
       # Bond center matrix and half bond vector
       #
        noa = self.noa
        self.Rab = full.matrix((noa, noa, 3))
        self.dRab = full.matrix((noa, noa, 3))
        for a in range(noa):
            for b in range(noa):
                self.Rab[a, b, :] = (self.R[a, :] + self.R[b, :])/2
                self.dRab[a, b, :] = (self.R[a, :] - self.R[b, :])/2

    @property
    def Rc(self):
#
# Form Rc  molecular gauge origin, default nuclear center of charge
#
        if self._Rc is not None:
            return self._Rc
        if self.gc is None:
            self._Rc = self.Z*self.R/self.Z.sum()
        else: 
            self._Rc = numpy.array(self.gc).view(full.matrix)
        return self._Rc


    @property
    def D(self):
        """ 
        Density from SIRIFC in blocked loprop basis
        """
        if self._D is not None:
            return self._D

        Di, Dv = dens.ifc(filename=self.sirifc)
        D = Di + Dv
        Ti = self.T.I
        self._D = ( Ti * D * Ti.T ).subblocked(self.cpa, self.cpa)
        return self._D

    @property
    def T(self):
        """
        Generate loprop transformation matrix according to the
        following steps 
        Given atomic overlap matrix:
        1. orthogonalize in each atomic block
        2. a) Lowdin orthogonalize occupied subspace
           b) Lowdin orthogonalize virtual subspace
        3. project occupied out of virtual
        4. Lowdin orthogonalize virtual
        Input: overlap S (matrix)
               contracted per atom (list)
               occupied per atom (nested list)
        Returns: transformation matrix T
                 such that T+ST = 1 (unit) """

        if self._T is not None: return self._T

        S = self.S()
        cpa = self.cpa
        opa = self.opa
    #
    # 1. orthogonalize in each atomic block
    #
        #t1=timing("step 1")
        if self.debug:#pragma: no cover
            print("Initial S", S)
        nbf = S.shape[0]
        #
        # obtain atomic blocking
        #
        #assert(len(cpa) == len(opa))
        noa = len(opa)
        nocc = 0
        for at in range(noa):
            nocc += len(opa[at])
        if self.debug:#pragma: no cover
            print("nocc", nocc)
        Satom = S.block(cpa, cpa)
        Ubl = full.unit(nbf).subblocked((nbf,), cpa)
        if self.debug:#pragma: no cover
            print("Ubl", Ubl)
        #
        # Diagonalize atomwise
        #
        T1 = blocked.BlockDiagonalMatrix(cpa, cpa)
        for at in range(noa):
            T1.subblock[at] = Ubl.subblock[0][at].GST(S)
        T1 = T1.unblock()
        if self.debug:#pragma: no cover
            print("T1", T1)
        #
        # Full transformration
        #
        S1 = T1.T * S * T1
        if self.debug:#pragma: no cover
            print("Overlap after step 1", S1)
        #t1.stop()
       
        # 2. a) Lowdin orthogonalize occupied subspace
        #
        # Reorder basis (permute)
        #
        #t2=timing("step 2")
        vpa = []
        adim = []
        for at in range(noa):
            vpa.append(cpa[at]-len(opa[at]))
            adim.append(len(opa[at]))
            adim.append(vpa[at])
        if self.debug:#pragma: no cover
            print("Blocking: Ao Av Bo Bv...", adim)
        #
        # dimensions for permuted basis
        #
        pdim = []
        if self.debug:#pragma: no cover
            print("opa", opa)
        for at in range(noa):
            pdim.append(len(opa[at]))
        for at in range(noa):
            pdim.append(vpa[at])
        if self.debug:#pragma: no cover
            print("Blocking: Ao Bo... Av Bv...", pdim)
        #
        # within atom permute occupied first
        #
        P1 = subblocked.matrix(cpa, cpa)
        for at in range(noa):
            P1.subblock[at][at][:, :] = full.permute(opa[at], cpa[at])
        n = len(adim)
        if self.debug:#pragma: no cover
            print("P1", P1)
        P1 = P1.unblock()
       
       
        P2 = subblocked.matrix(adim, pdim)
        for i in range(0, len(adim), 2):
            P2.subblock[i][i//2] = full.unit(adim[i])
        for i in range(1, len(adim), 2):
            P2.subblock[i][noa+(i-1)//2] = full.unit(adim[i])
        if self.debug:#pragma: no cover
            print("P2", P2)
        P2 = P2.unblock()
       
        #
        # new permutation scheme
        #
       
        P = P1*P2
        if self.debug:#pragma: no cover
            print("P", P)
            if not numpy.allclose(P.inv(), P.T):
                print("P not unitary")
                sys.exit(1)
       
       
        S1P = P.T*S1*P
        if self.debug:#pragma: no cover
            print("Overlap in permuted basis", S1P)
       
       
        #invsq=lambda x: 1.0/math.sqrt(x)
        occdim = (nocc, sum(vpa))
        S1Pbl = S1P.block(occdim, occdim)
        ### SYM ### S1Pbl += S1Pbl.T; S1Pbl *= 0.5 ###SYM###
        #T2bl=S1Pbl.func(invsq)
        T2bl = S1Pbl.invsqrt()
        T2 = T2bl.unblock()
       
       
        S2 = T2.T*S1P*T2
        if self.debug:#pragma: no cover
            print("Overlap after step 2", S2)
        #t2.stop()
       
        #
        # Project occupied out of virtual
        #
        #t3=timing("step 3")
        if 0:
            T3 = full.unit(nbf).GST(S2)
        else:
            S2sb = S2.subblocked(occdim, occdim)
            T3sb = full.unit(nbf).subblocked(occdim, occdim)
            T3sb.subblock[0][1] = -S2sb.subblock[0][1]
            T3 = T3sb.unblock()
        S3 = T3.T*S2*T3
        #
        if self.debug:#pragma: no cover
            print("T3", T3)
            print("Overlap after step 3", S3)
        #t3.stop()
        #
        # 4. Lowdin orthogonalize virtual 
        #
        #t4=timing("step 4")
        T4b = blocked.unit(occdim)
        S3b = S3.block(occdim, occdim)
        if self.debug:#pragma: no cover
            print("S3b", S3b)
            print("T4b", T4b)
        ### SYM ### S3b += S3b.T; S3b *= 0.5 ###SYM###
        T4b.subblock[1] = S3b.subblock[1].invsqrt()
        T4 = T4b.unblock()
        S4 = T4.T*S3*T4
        #S4=S3
        if self.debug:#pragma: no cover
            print("T4", T4)
            print("Overlap after step 4", S4)
        #t4.stop()
       
        #
        # permute back to original basis
        #
        S4 = P*S4*P.T
        if self.debug:#pragma: no cover
            print("Final overlap ", S4)
       
        #
        # Return total transformation
        #
        T = T1*P*T2*T3*T4*P.T
        #
        # Test
        #
        if self.debug:#pragma: no cover
            print("Transformation determinant", T.det())
            print("original S", S, "final", T.T*S*T)
        self._T = T
        return self._T

    @property
    def Qab(self):
        """ set charge/atom property"""
        if self._Qab is not None: return self._Qab

        D = self.D
        T = self.T
        cpa = self.cpa
        
        noa = self.noa
        _Qab = full.matrix((noa, noa))
        for a in range(noa):
            _Qab[a, a] = - D.subblock[a][a].tr()
        self._Qab = _Qab
        return self._Qab

    @property 
    def Qa(self):
        return self.Qab.diagonal()

    @property
    def Dab(self):
        """Set dipole property"""

        if self._Dab is not None: return self._Dab

        x = self.x
        D = self.D
        Rab = self.Rab
        Qab = self.Qab
       
        noa = self.noa
        _Dab = full.matrix((3, noa, noa))
        for i in range(3):
            for a in range(noa):
                for b in range(noa):
                    _Dab[i, a, b] = -(
                         x[i].subblock[a][b]&D.subblock[a][b]
                         ) \
                         -Qab[a, b]*Rab[a, b, i]
        
        self._Dab = _Dab
        return self._Dab

    @property
    def Da(self):
        """Sum up bonds contributions to atom"""
        if self._Da is not None: return self._Da

        Dab = self.Dab
        self._Da = Dab.sum(axis=2).view(full.matrix)
        return self._Da

    @property
    def Dsym(self):
        """Symmetrize density contributions from atom pairs """
        if self._Dsym is not None: return self._Dsym

        Dab = self.Dab
        noa = self.noa
        dsym = full.matrix((3, noa*(noa+1)//2))
        ab = 0
        for a in range(noa):
            for b in range(a):
                dsym[:, ab] = Dab[:, a, b] + Dab[:, b, a]
                ab += 1
            dsym[:, ab] = Dab[:, a, a]
            ab += 1

        self._Dsym = dsym
        return self._Dsym

    @property
    def Dtot(self):
        _Dtot = self.Da.sum(axis=1).view(full.matrix) 
        _Dtot += self.Qa*self.R - self.Qa.sum()*self.Rc
        return _Dtot

    @property
    def QUab(self):
        """Quadrupole moment"""
        if self._QUab is not None: return self._QUab

        D = self.D
        R = self.R
        Rc = self.Rc
        dRab = self.dRab
        Qab = self.Qab
        Dab = self.Dab

        lab = ("XXSECMOM", "XYSECMOM", "XZSECMOM", 
                           "YYSECMOM", "YZSECMOM", 
                                       "ZZSECMOM")

        xy = self.getprop(*lab)

        noa = self.noa
        QUab = full.matrix((6, noa, noa))
        rrab = full.matrix((6, noa, noa))
        rRab = full.matrix((6, noa, noa))
        RRab = full.matrix((6, noa, noa))
        Rab = self.Rab
        for a in range(noa):
            for b in range(noa):
                ij = 0
                for i in range(3):
                    for j in range(i, 3):
                        rrab[ij, a, b] = -(
                            xy[ij].subblock[a][b]&D.subblock[a][b]
                            ) 
                        rRab[ij, a, b] = Dab[i, a, b]*Rab[a,b,j]+Dab[j, a, b]*Rab[a,b,i]
                        RRab[ij, a, b] = Rab[a,b,i]*Rab[a,b,j]*Qab[a, b]
                        ij += 1
        QUab = rrab-rRab-RRab
        self._QUab = QUab
        #
        # Addition term - gauge correction summing up bonds
        #
        dQUab = full.matrix(self.QUab.shape)
        for a in range(noa):
            for b in range(noa):
                ij = 0
                for i in range(3):
                    for j in range(i, 3):
                        dQUab[ij, a, b] = dRab[a, b, i]*Dab[j, a, b] \
                                      +dRab[a, b, j]*Dab[i, a, b]
                        ij += 1
        self.dQUab = - dQUab

        return self._QUab

    @property
    def QUa(self):
        """Sum up quadrupole bond terms to atoms"""
        if self._QUa is not None: return self._QUa

        QUab = self.QUab + self.dQUab
        noa = self.noa
        self._QUa = QUab.sum(axis=2).view(full.matrix)
        return self._QUa

    @property
    def QUsym(self):
        """Quadrupole moment symmetrized over atom pairs"""
        if self._QUsym is not None: return self._QUsym

        QUab = self.QUab
        noa = self.noa
        qusym = full.matrix((6, noa*(noa+1)//2))
        ab = 0
        for a in range(noa):
            for b in range(a):
                qusym[:, ab] = QUab[:, a, b] + QUab[:, b, a]
                ab += 1
            qusym[:, ab] = QUab[:, a, a]
            ab += 1

        self._QUsym = qusym
        return self._QUsym

    @property
    def QUN(self):
        """Nuclear contribution to quadrupole"""
        if self._QUN is not None: return self._QUN

        qn = full.matrix(6)
        Z = self.Z
        R = self.R
        Rc = self.Rc
        for a in range(len(Z)):
            ij = 0
            for i in range(3):
                for j in range(i, 3):
                    qn[ij] += Z[a]*(R[a, i]-Rc[i])*(R[a, j]-Rc[j])
                    ij += 1
        self._QUN = qn
        return self._QUN

    @property
    def QUc(self):
        if self._QUc is not None: return self._QUc

        rrab=full.matrix((6, self.noa, self.noa))
        rRab=full.matrix((6, self.noa, self.noa))
        RRab=full.matrix((6, self.noa, self.noa))
        Rabc = 1.0*self.Rab
        for a in range(self.noa):
            for b in range(self.noa):
                Rabc[a,b,:] -= self.Rc
        for a in range(self.noa):
            for b in range(self.noa):
                ij = 0
                for i in range(3):
                    for j in range(i,3):
                        rRab[ij, a, b] = self.Dab[i, a, b]*Rabc[a, b, j]\
                                       + self.Dab[j, a, b]*Rabc[a, b, i]
                        RRab[ij, a, b] = self.Qab[a, b]*(self.R[a, i] - self.Rc[i])*(self.R[b, j] - self.Rc[j])
                        ij += 1
        QUcab = self.QUab + rRab + RRab
        self._QUc = QUcab.sum(axis=2).sum(axis=1).view(full.matrix)
        return self._QUc

    @property
    def Fab(self, **kwargs):
        """Penalty function"""
        if self._Fab is not None: return self._Fab

        Fab = full.matrix((self.noa, self.noa))
        for a in range(self.noa):
            Za = self.Z[a]
            Ra = self.R[a]
            for b in range(a):
                Zb = self.Z[b]
                Rb = self.R[b]
                Fab[a, b] = self.pf(Za, Ra, Zb, Rb, **kwargs)
                Fab[b, a] = Fab[a, b]
        for a in range(self.noa):
            Fab[a, a] += - Fab[a, :].sum()
        self._Fab = Fab
        return self._Fab

    @property
    def la(self):
        """Lagrangian for local poplarizabilities"""
        #
        # The shift should satisfy
        #   sum(a) sum(b) (F(a,b) + C)l(b) = sum(a) dq(a) = 0
        # =>sum(a, b) F(a, b) + N*C*sum(b) l(b) = 0
        # => C = -sum(a, b)F(a,b) / sum(b)l(b)
        #

        if self._la is not None: return self._la
        #
        dQa = self.dQa
        Fab = self.Fab
        Lab = Fab + self.sf(Fab)
        self._la = [rhs/Lab for rhs in dQa]
        return self._la

    @property
    def l2a(self):
        """Lagrangian for local poplarizabilities"""
        #
        # The shift should satisfy
        #   sum(a) sum(b) (F(a,b) + C)l(b) = sum(a) dq(a) = 0
        # =>sum(a, b) F(a, b) + N*C*sum(b) l(b) = 0
        # => C = -sum(a, b)F(a,b) / sum(b)l(b)
        #

        if self._l2a is not None: return self._l2a
        #
        d2Qa = self.d2Qa
        Fab = self.Fab
        Lab = Fab + self.sf(Fab)
        self._l2a = [rhs/Lab for rhs in d2Qa]
        return self._l2a

    @property
    def Dk(self):
        """Read perturbed densities"""

        if self._Dk is not None:
            return self._Dk

        lab = ['XDIPLEN', "YDIPLEN", "ZDIPLEN"]
        prp = os.path.join(self.tmpdir,"AOPROPER")
        T = self.T
        cpa = self.cpa

        Dkao = lr.Dk(*lab, freqs=self.freqs, tmpdir=self.tmpdir, absorption=self.damping, lr_vecs=True)
        if self.damping:
            Re_Dkao, Im_Dkao = Dkao
            if self.real_pol:
                _Dk = {lw:(T.I*Re_Dkao[lw]*T.I.T).subblocked(cpa, cpa) for lw in Re_Dkao}
            elif self.imag_pol:
                _Dk = {lw:(T.I*Im_Dkao[lw]*T.I.T).subblocked(cpa, cpa) for lw in Im_Dkao}
            else:
                raise ValueError
        else:
            _Dk = {lw:(T.I*Dkao[lw]*T.I.T).subblocked(cpa, cpa) for lw in Dkao}

        self._Dk = _Dk
        return self._Dk

    @property
    def D2k(self):
        """Read perturbed densities"""

        if self._D2k is not None:
            return self._D2k

        lab = ['XDIPLEN ', "YDIPLEN ", "ZDIPLEN "]
        qrlab = [lab[j]+lab[i] for i in range(3) for j in range(i,3)]
        prp = os.path.join(self.tmpdir, "AOPROPER")
        T = self.T
        cpa = self.cpa

        Dkao = qr.D2k(*qrlab, freqs=self.freqs, tmpdir=self.tmpdir)
        #print("Dkao.keys", Dkao.keys())
        _D2k = {lw:(T.I*Dkao[lw]*T.I.T).subblocked(cpa, cpa) for lw in Dkao}

        self._D2k = _D2k
        return self._D2k

    @property
    def x(self):
        """Read dipole matrices to blocked loprop basis"""

        if self._x is not None:
            return self._x
        
        lab = ['XDIPLEN', "YDIPLEN", "ZDIPLEN"]

        self._x = self.getprop(*lab) 
        return self._x

    def getprop(self, *args):
        """Read general property matrices to blocked loprop basis"""

        T = self.T
        cpa = self.cpa
        prp = os.path.join(self.tmpdir,"AOPROPER")

        return [
            (T.T*p*T).subblocked(cpa, cpa) for p in
            prop.read(*args, filename=prp, unpack=True)
            ]


    @property
    def dQa(self):
        """Charge shift per atom"""
        if self._dQa is not None: return self._dQa

        T = self.T
        cpa = self.cpa
        noa = self.noa

        Dk = self.Dk
        labs = ('XDIPLEN', 'YDIPLEN', 'ZDIPLEN')

        dQa = full.matrix((self.nfreqs, noa, 3))
        for a in range(noa):
            for il, l in enumerate(labs):
                for iw, w in enumerate(self.freqs):
                    dQa[iw, a, il] = - Dk[(l,w)].subblock[a][a].tr()
        self._dQa = dQa
        return self._dQa

    @property
    def d2Qa(self):
        """Charge shift per atom"""
        if self._d2Qa is not None: return self._d2Qa

        T = self.T
        cpa = self.cpa
        noa = self.noa

        D2k = self.D2k
                

        # static
        wb = wc = 0.0
        d2Qa = full.matrix((1, noa, 6))

        lab = ['XDIPLEN ', "YDIPLEN ", "ZDIPLEN "]
        qrlab = [lab[j]+lab[i] for i in range(3) for j in range(i,3)]

        for a in range(noa):
            for il, l in enumerate(qrlab):
                il = qrlab.index(l)
                d2Qa[0, a, il] = - D2k[(l,wb,wc)].subblock[a][a].tr()
        self._d2Qa = d2Qa
        return self._d2Qa

    @property
    def dQab(self):
        """Charge transfer matrix"""
        if self._dQab is not None: return self._dQab

        dQa = self.dQa
        la = self.la
        noa = self.noa

        dQab = full.matrix((self.nfreqs, noa, noa, 3))
        for field in range(3):
            for a in range(noa):
                Za = self.Z[a]
                Ra = self.R[a]
                for b in range(a):
                    Zb = self.Z[b]
                    Rb = self.R[b]
                    for w in self.rfreqs:
                        dQab[w, a, b, field] =  \
                        - (la[w][a, field]-la[w][b, field]) * \
                        self.pf(Za, Ra, Zb, Rb)
                        dQab[w, b, a, field] = -dQab[w, a, b, field]

        self._dQab = dQab
        return self._dQab

    @property
    def d2Qab(self):
        """Charge transfer matrix for double perturbation"""
        if self._d2Qab is not None: return self._d2Qab

        d2Qa = self.d2Qa
        l2a = self.l2a
        noa = self.noa

        d2Qab = full.matrix((self.nfreqs, noa, noa, 6))
        for field in range(6):
            for a in range(noa):
                Za = self.Z[a]
                Ra = self.R[a]
                for b in range(a):
                    Zb = self.Z[b]
                    Rb = self.R[b]
                    for w in self.rfreqs:
                        d2Qab[w, a, b, field] =  \
                        - (l2a[w][a, field]-l2a[w][b, field]) * \
                        self.pf(Za, Ra, Zb, Rb)
                        d2Qab[w, b, a, field] = -d2Qab[w, a, b, field]

        self._d2Qab = d2Qab
        return self._d2Qab


    @property
    def Aab(self):
        """Localized polariziabilities:

        Contribution from change in localized dipole moment
        - d (r - R(AB)):D(AB) = - r:dD(AB) + dQ(A) R(A) \delta(A,B)
        """
        if self._Aab is not None: return self._Aab

        D = self.D
        Dk = self.Dk
        #T = self.T
        cpa = self.cpa
        Z = self.Z
        Rab = self.Rab
        Qab = self.Qab
        dQa = self.dQa
        x = self.x

        noa = len(cpa)
        labs = ('XDIPLEN', 'YDIPLEN', 'ZDIPLEN')
        Aab = full.matrix((self.nfreqs, 3, 3, noa, noa))

        # correction term for shifting origin from O to Rab
        for i,li in enumerate(labs):
            for j,lj in enumerate(labs):
                for a in range(noa):
                    for b in range(noa):
                        for jw, w in enumerate(self.freqs):
                            Aab[jw, i, j, a, b] = (
                           -x[i].subblock[a][b]&Dk[(lj, w)].subblock[a][b]
                           )
                    for jw in self.rfreqs:
                        Aab[jw, i, j, a, a] -= dQa[jw, a, j]*Rab[a, a, i]

        self._Aab = Aab
        return self._Aab


    @property
    def dAab(self):
        """Charge transfer contribution to bond polarizability"""
        if self._dAab is not None: return self._dAab

        dQa = self.dQa
        dQab = self.dQab
        dRab = self.dRab
        noa = self.noa
        dAab = full.matrix((self.nfreqs, 3, 3, noa, noa))
        for a in range(noa):
            for b in range(noa):
                for i in range(3):
                    for j in range(3):
                        dAab[:, i, j, a, b] = (
                            dRab[a, b, i]*dQab[:, a, b, j]+
                            dRab[a, b, j]*dQab[:, a, b, i]
                            )
        self._dAab = dAab
        return self._dAab


    @property
    def Am(self):
        """Molecular polarizability:

        To reconstruct the molecular polarizability  from localized 
        polarizabilties one has to reexpand in terms of an arbitrary but common 
        origin leading to the correction term below

        d<-r> = - sum(A,B) (r-R(A,B))dD(A,B) + R(A) dQ(A) \delta(A,B)
        """

        if self._Am is not None: return self._Am

        dQa = self.dQa
        Rab = self.Rab
        Aab = self.Aab
        dAab = self.dAab
        noa = self.noa

        self._Am = (Aab + 0.5*dAab).sum(axis=4).sum(axis=3).view(full.matrix)
        return self._Am

    @property
    def Bab(self):
        """Localized hyperpolariziabilities"""
        if self._Bab is not None: return self._Bab

        D = self.D
        D2k = self.D2k
        #T = self.T
        cpa = self.cpa
        Z = self.Z
        Rab = self.Rab
        Qab = self.Qab
        d2Qa = self.d2Qa
        x = self.x

        noa = len(cpa)
        labs = ('XDIPLEN ', 'YDIPLEN ', 'ZDIPLEN ')
        qlabs = [labs[i] + labs[j] for i in range(3) for j in range(i,3)]
        Bab = full.matrix( (self.nfreqs, 3, 6, noa, noa) )

        #pdb.set_trace()

        #correction term for shifting origin from O to Rab
        for i, li in enumerate(labs):
            for jk,ljk in enumerate(qlabs):
                #print(i,jk, li, ljk)
                for a in range(noa):
                    for b in range(noa):
                        for iw, w in enumerate(self.freqs):
                            Bab[iw, i, jk, a, b] = (
                                -x[i].subblock[a][b] & D2k [(ljk, w, w)].subblock[a][b]
                                )
                    for iw in self.rfreqs:
                        Bab[iw, i, jk, a, a] -= d2Qa[iw, a, jk]*Rab[a, a, i]

        self._Bab = Bab
        return self._Bab

    @property
    def dBab(self):
        """Charge transfer contribution to bond hyperpolarizabilitypolarizability"""
        if self._dBab is not None: return self._dBab

        dQa = self.dQa
        d2Qa = self.d2Qa
        dQab = self.dQab
        d2Qab = self.d2Qab
        dRab = self.dRab
        noa = self.noa
        dBab = full.matrix((self.nfreqs, 3, 6, noa, noa))
        for a in range(noa):
            for b in range(noa):
                for i in range(3):
                    for j in range(6):
                        dBab[:, i, j, a, b] = 2*dRab[a, b, i]*d2Qab[:, a, b, j]
        self._dBab = dBab
        return self._dBab

    @property
    def Bm(self):
        "Molecular hyperpolarizability"

        if self._Bm is not None: return self._Bm

        d2Qa = self.d2Qa
        Rab = self.Rab
        Bab = self.Bab #+ 0.25 * self.dBab
        dBab = self.dBab
        noa = self.noa

        self._Bm = (Bab + 0.5*dBab).sum(axis=4).sum(axis=3).view(full.matrix)

        return self._Bm

    def output_by_atom(self, fmt="%9.5f", max_l=0, pol=0, hyperpol=0, bond_centers=False, angstrom=False):
        """Print nfo"""

        if max_l >= 0: 
            Qab = self.Qab
            Qa = Qab.diagonal()
        if max_l >= 1:
            Dab = self.Dab
            Da = self.Da
            Dsym = self.Dsym
        if max_l >= 2:
            QUab = self.QUab
            QUN = self.QUN
            dQUab = self.dQUab
            QUa = self.QUa
        if  pol:
            Aab = self.Aab + self.dAab
        if  hyperpol:
            Bab = self.Bab + self.dBab
         
        if angstrom: 
            unit = "AA" 
            xconv = 0.5291772108
            xconv3 = 0.5291772108**3
        else: 
            unit = "AU"
            xconv = 1
            xconv3 = 1

        Z = self.Z
        R = self.R
        Rc = self.Rc
        noa = self.noa
    #
    # Form net atomic properties P(a) = sum(b) P(a,b)
    #
        if self._Aab is not None: 
            Aab = self.Aab + 0.5*self.dAab
            Aa = Aab.sum(axis=4)

        if self._Bab is not None: 
            Bab = self.Bab + 0.5*self.dBab
            Ba = Bab.sum(axis=4)

        if bond_centers:
            for a in range(noa):
                for b in range(a):
                    header("Bond    %d %d" % (a+1, b+1))
                    print("Bond center:       " + \
                        (3*fmt) % tuple(0.5*(R[a, :]+R[b, :])*xconv)
                        )
                    if max_l >= 0:
                        print("Electronic charge:   "+fmt % Qab[a, b])
                        print("Total charge:        "+fmt % Qab[a, b])
                    if self._Dab is not None:
                        print("Electronic dipole    " + \
                            (3*fmt) % tuple(Dab[:, a, b]+Dab[:, b, a]))
                        print("Electronic dipole norm" + \
                            fmt % (Dab[:, a, b]+Dab[:, b, a]).norm2()
                            )
                    if self._QUab is not None:
                        raise NotImplementedError
#                       print("Electronic quadrupole" + \
#                           (6*fmt) % tuple(QUab[:, a, b]+QUab[:, b, a]))
                    if self._Aab is not None:
                        for iw, w in enumerate(self.freqs):
                            Asym = Aab[iw, :, :, a, b] + Aab[iw, :, :, b, a]
                            if pol > 0:
                                print("Isotropic polarizability (%g)" % w + fmt % (Asym.trace()/3*xconv3))
                            if pol > 1:
                                print("Polarizability (%g)      " % w)
                                print((6*fmt) % tuple(Asym.pack().view(full.matrix)*xconv3))

                    if self._Bab is not None:
                        for iw, w in enumerate(self.freqs):
                            Bsym = Bab[iw, :, :, a, b] + Bab[iw, :, :, b, a]
                            output_beta(Bsym, self.Da[:, a])

                header("Atom    %d"%(a+1))
                print("Atom center:       " + \
                    (3*fmt) % tuple(R[a,:]*xconv))
                print("Nuclear charge:    "+fmt % Z[a])
                if max_l >= 0:
                    print("Electronic charge:   "+fmt % Qab[a, a])
                    print("Total charge:        "+fmt % (Z[a]+Qab[a, a]))
                if self._Dab is not None:
                    print("Electronic dipole    " + \
                        (3*fmt) % tuple(Dab[:, a, a]))
                    print("Electronic dipole norm" + \
                        fmt % Dab[:, a, a].norm2())
                if self._QUab is not None:
                    print("Electronic quadrupole" + \
                        (6*fmt) % tuple(QUab[:, a, a]))
                if self._Aab is not None:
                    for iw, w in enumerate(self.freqs):
                        Asym = Aab[iw, :, :, a, a] 
                        if pol > 0:
                            print("Isotropic polarizability (%g)" % w + fmt % (Asym.trace()/3*xconv3))
                        if pol > 1:
                            print("Polarizability (%g)       " % w + (6*fmt) % tuple(Asym.pack().view(full.matrix)*xconv3))
                if self._Bab is not None:
                    for iw, w in enumerate(self.freqs):
                        Bsym = Bab[iw, :, :, a, a] 
                        output_beta(Bsym, Da[:, a])

        else:
            for a in range(noa):
                header("Atomic domain %d" % (a+1))
                print("Domain center:       "+(3*fmt) % tuple(R[a, :]*xconv))
                line = " 0"
                line += (3*"%17.10f") % tuple(xtang*R[a, :])
                print("Nuclear charge:      "+fmt % Z[a])
                if self.max_l >= 0:
                    print("Electronic charge:   "+fmt % Qa[a])
                    print("Total charge:        "+fmt % (Z[a]+Qa[a]))
                if self._Dab is not None:
                    print("Electronic dipole    "+(3*fmt) % tuple(self.Da[:, a]))
                    print("Electronic dipole norm"+(fmt) % self.Da[:, a].view(full.matrix).norm2())
                if self._QUab is not None:
                    #print("QUab", QUab)
                    print("Electronic quadrupole"+(6*fmt) % tuple(QUa[:, a]))
                if self._Aab is not None:
                    for iw, w in enumerate(self.freqs):
                        Asym = Aa[iw, :, :, a].view(full.matrix)
                        print("Isotropic polarizablity (w=%g)" % w + fmt % (Aa[iw, :, :, a].trace()/3*xconv3))
                        print("Electronic polarizability (w=%g)" % w + \
                            (6*fmt) % tuple(Asym.pack().view(full.matrix)*xconv3))
                if self._Bab is not None:
                    for iw, w in enumerate(self.freqs):
                        Bsym = Ba[iw, :, :, a].view(full.matrix)
                        output_beta(Bsym, self.Da[:, a])
    #
    # Total molecular properties
    #
        Ztot = Z.sum()
        if self.max_l >= 0:
            Qtot = Qa.sum()
        if self.max_l >= 1:
            Dm = self.Da.sum(axis=1).view(full.matrix)
            Dc = Qa*(R-Rc)
            DT = Dm+Dc
        if self._QUab is not None:
            QUm = self.QUc
            QUT = QUm+QUN
        if self._Bab is not None:
            Dm = self.Da.sum(axis=1).view(full.matrix)

        header("Molecular")
        print("Domain center:       "+(3*fmt) % tuple(Rc*xconv))
        print("Nuclear charge:      "+fmt % Ztot)

        if self.max_l >= 0:
            print("Electronic charge:   "+fmt % Qtot)
            print("Total charge:        "+fmt % (Ztot+Qtot))

        if self.max_l >= 1:
            print("Electronic dipole    "+(3*fmt) % tuple(Dm))
            print("Gauge   dipole       "+(3*fmt) % tuple(Dc))
            print("Total   dipole       "+(3*fmt) % tuple(DT))

        if self._QUab is not None:
            print("Electronic quadrupole"+(6*fmt) % tuple(QUm))
            print("Nuclear    quadrupole"+(6*fmt) % tuple(QUN))
            print("Total      quadrupole"+(6*fmt) % tuple(QUT))

        if self._Aab is not None:
            for iw, w in enumerate(self.freqs):
                Am = self.Am[iw]
                print("Polarizability av (%g)   " % w + fmt % (Am.trace()/3*xconv3))
                print("Polarizability (%g)      " % w + (6*fmt) % tuple(Am.pack().view(full.matrix)*xconv3))

        if self._Bab is not None:
            for iw, w in enumerate(self.freqs):
                Bm = self.Bm[iw]
                output_beta(Bm, dip=Dm, fmt=fmt)

    def output_template(self, maxl = 0, pol = 0, hyper = 0, template_full = False, decimal = 4, full_loc =0, freqs = None):
        l_dict = { 0 : "charge", 1 : "dipole", 2 : "quadrupole",
                }
#Upper triangular alpha
        a_dict = { 0 : "", 2 : "alpha" }
#Upper triangular beta
        b_dict = { 0 : "", 2 : "beta" }

        fmt = "%."+"%df" %decimal
        line = ""

        if pol > 0:
            Aab = self.Aab + 0.5 * self.dAab
        if hyper > 0:
            Bab = self.Bab + 0.5 * self.dBab


        if maxl not in l_dict:
            print("ERROR: called output_template with wrong argument range")
        if pol not in a_dict:
            print("ERROR: called output_template with wrong argument range")
        if hyper not in b_dict:
            print("ERROR: called output_template with wrong argument range")

        elem_dict = {1:"H", 6:"C", 7: "N", 8 : "O", 16 : "S"}
        if maxl >= 0:
            if template_full:
# Put point dipole on center of charge
                line += "( '%s%d', "%(elem_dict[self.Z[full_loc]] , full_loc + 1) + '"charge") : [ %s ],\n'%fmt %(self.Z.sum()+ self.Qab.sum())
            else:
                for a in range(self.noa):
                    line += "( '%s%d', "%(elem_dict[self.Z[a]] ,a+1) + '"charge") : [ %s ],\n'%fmt %(self.Z[a] + self.Qab[a, a])
        if maxl >= 1:
            if template_full:
                Dm = self.Da.sum(axis=1).view(full.matrix)
                Dc = self.Qab.diagonal()*(self.R-self.Rc)
                DT = Dm+Dc
                line += "( '%s%d', "%(elem_dict[self.Z[full_loc]] , full_loc + 1) + '"dipole") : [ %s, %s, %s ],\n'%tuple([fmt for i in range(3)]) %(tuple(DT))
            else:
                for a in range(self.noa):
                    line += "( '%s%d', "%(elem_dict[self.Z[a]] ,a+1) + '"dipole") : [ %s, %s, %s ],\n'%tuple([fmt for i in range(3)]) %(tuple(self.Dab.sum(axis=2)[:, a]))
        if maxl >= 2:
            if template_full:
                line += "( '%s%d', "%(elem_dict[self.Z[full_loc]] , full_loc + 1) +  '"quadrupole") : [ %s, %s, %s, %s, %s, %s ],\n'%tuple([fmt for i in range(6)]) %(tuple((self.QUab+self.dQUab).sum(axis=(1,2))[:]))
            else:
                for a in range(self.noa):
                    line +=  "( '%s%d', "%(elem_dict[self.Z[a]] ,a+1) + '"quadrupole") : [ %s, %s, %s, %s, %s, %s ],\n'%tuple([fmt for i in range(6)]) %(tuple((self.QUab+self.dQUab).sum(axis=2)[:, a]))
        if pol >= 2:
            if template_full:
                Asym = Aab.sum(axis=(3,4))[0, :, :].view(full.matrix)
                A = Asym.pack().view(full.matrix).copy()
                A[2], A[3] = A[3], A[2]
                line += "( '%s%d', "%(elem_dict[self.Z[full_loc]] , full_loc + 1) + '"alpha") : [ %s, %s, %s, %s, %s, %s ],\n'%tuple([fmt for i in range(6)]) %(tuple(A))
            else:
                for a in range(self.noa):
# Only for one frequency for now, todo, fix later if needed general
                    Asym = Aab.sum(axis=4)[0, :, :, a].view(full.matrix)
                    A = Asym.pack().view(full.matrix).copy()
                    A[2], A[3] = A[3], A[2]
                    line += "( '%s%d', "%(elem_dict[self.Z[a]] ,a+1) +  '"alpha") : [ %s, %s, %s, %s, %s, %s ],\n'%tuple([fmt for i in range(6)]) %(tuple(A))
        if hyper >= 2:
            if template_full:
                Bsym = symmetrize_first_beta( Bab.sum(axis=(3,4))[0, :, :].view(full.matrix) )
                line += "( '%s%d', "%(elem_dict[self.Z[full_loc]] , full_loc + 1) + '"beta") : [ %s, %s, %s, %s, %s, %s, %s, %s, %s, %s ],\n' %tuple([fmt for i in range(len(Bsym))]) %(tuple(Bsym))
            else:
                for a in range(self.noa):
# Only for one frequency for now, todo, fix later if needed general
                    Bsym = symmetrize_first_beta( Bab.sum(axis=4)[0, :, :, a].view(full.matrix) )
                    line += "( '%s%d', "%(elem_dict[self.Z[a]] ,a+1) +  '"beta") : [ %s, %s, %s, %s, %s, %s, %s, %s, %s, %s ],\n' %(tuple([fmt for i in range(len(Bsym))])) %(tuple(Bsym))
        return line



    def output_potential_file(
            self, maxl, pol, hyper, bond_centers=False, angstrom=False, decimal = 3,
            ):
        """Output potential file"""
        fmt = "%" + "%d." %(7 + decimal) + "%df" % decimal
        lines = []
        if angstrom: 
            unit = "AA" 
            xconv = 0.5291772108
            xconv3 = 0.5291772108**3
        else: 
            unit = "AU"
            xconv = 1
            xconv3 = 1
        lines.append(unit)


        noa = self.noa

#To get number of centers and bonding is on
        bond_mat = numpy.zeros( (noa, noa,), dtype = int )
        for a in range( noa ):
            for b in range( a ):
                r = numpy.sqrt( (( self.R[a] - self.R[b])**2 ).sum() )
                if r < bond_co[ (int(self.Z[a]), int(self.Z[b])) ]/xtang:
                    bond_mat[ a, b ] = 1
                    bond_mat[ b, a ] = 1

        if bond_centers:
#Where the number of bonds is the diagonal plus each entry with '1'
#in the upper triangular of bond_mat
            noc = bond_mat.shape[0] + reduce(lambda a,x: a + len(numpy.where(x==1)[0]), [row[i+1:] for i, row in enumerate(bond_mat)], 0 )
            #noc = noa*(noa + 1)/2
        else:
            noc = self.noa

        lines.append("%d %d %d %d"%(noc, maxl, pol, 1))

        if maxl >= 0: Qab = self.Qab
        if maxl >= 1: 
            Dab = self.Dab
            Dsym = self.Dsym
        if maxl >= 2: 
            QUab = self.QUab
            dQUab = self.dQUab
        if pol > 0:
            Aab = self.Aab + 0.5*self.dAab

        if hyper > 0:
            Bab = self.Bab + 0.5*self.dBab

        if bond_centers:
            ab = 0
            for a in range(noa):
                for b in range(a):
                    if bond_mat[ a, b]:
                        line  = ("1" + 3*fmt) % tuple(self.Rab[a, b, :])
                        if maxl >= 0:
                            line += fmt % Qab[a, b]
                        if maxl >= 1:
                            line += (3*fmt) % tuple( Dab[:,b,a] + Dab[:, a, b] )
                        if maxl >= 2: line += (6*fmt) % \
                            tuple(QUab[:, a, b] +QUab[:, b, a])
                        if pol > 0:
                            Aab = self.Aab + 0.5*self.dAab
                            for iw, w in enumerate(self.freqs):
                                Asym = Aab[iw, :, :, a, b] + Aab[iw, :, :, b, a]
                                if pol == 1: line += fmt % (Asym.trace()*xconv3/3)
                                elif (pol%10) == 2: 
                                    out = Asym.pack().view(full.matrix)*xconv3
                                    out[2:4] = out[3:1:-1]
                                    line += (6*fmt)%tuple(out )
                        if hyper > 0:
                            for iw, w in enumerate(self.freqs):
                                Bsym = Bab[iw, :, :, a, b] + Bab[iw, :, :, b, a]
                                #if hyper == 1:
                                #    dip = self.Da[:,a]
                                #    betakk = Bsym[:,0] + Bsym[:, 3] + Bsym[:, 5]
                                #    line += fmt % ( 0.2 * (betakk & dip) / dip.norm2() )
                                Btotsym = symmetrize_first_beta( Bsym )
                                line += 10*fmt % tuple( Btotsym )
                        lines.append(line)
#For atom a, non_bond_pos holds atoms that are not bonded to a
#Include only non bonded to atomic prop here 
                nbond_pos = numpy.where( bond_mat[ a ] == 0 )[0]

                line  = ("1" + 3*fmt) % tuple(self.Rab[a, a, :])
                if maxl >= 0:
                    line += fmt % (self.Z[a]+Qab[a, a])
                if maxl >= 1:
                    line += (3*fmt) % tuple( reduce(lambda x,y: x+ Dab[:, a, y], nbond_pos, 0.0 ))
                if maxl >= 2:
                    raise NotImplementedError
                if pol > 0:
                    for iw, w in enumerate(self.freqs):
                        if pol %10 == 2:
                            out = reduce( lambda x,y: x + Aab[iw, :, :, a, y], nbond_pos, 0.0 ).pack().view(full.matrix)*xconv3
                            out[2:4] = out[3:1:-1]
                            line += (6*fmt) % tuple(out)
                        elif pol == 1:
                            out = reduce( lambda x,y: x + Aab[iw, :, :, a, y],nbond_pos, 0.0 ).view(full.matrix).trace()/3.0 * xconv3
                            line += fmt % out
                if hyper > 0:
                    for iw, w in enumerate(self.freqs):
                        Bsym = reduce( lambda x,y: x + Bab[iw, :, :, a, y], nbond_pos, 0.0 ).view(full.matrix)
                        #if hyper == 1:
                        #    dip = self.Da[:,a]
                        #    betakk = Bsym[:,0] + Bsym[:, 3] + Bsym[:, 5]
                        #    line += fmt % ( 0.2 * (betakk & dip) / dip.norm2() )
                        Btotsym = symmetrize_first_beta( Bsym )
                        line += 10*fmt % tuple( Btotsym )
                ab += 1
                lines.append(line)
        else:
            for a in range(noa):
                line  = ("1" + 3*fmt) % tuple(self.Rab[a, a, :]*xconv)
                if maxl >= 0: line += fmt % (self.Z[a] + Qab[a, a])
                if maxl >= 1: line += (3*fmt) % tuple(Dab.sum(axis=2)[:, a])
                if maxl >= 2: 
                    line += (6*fmt) % tuple((QUab+dQUab).sum(axis=2)[:,  a])
                if pol > 0:
                    for iw in range(self.nfreqs):
                        Asym = Aab.sum(axis=4)[iw, :, :, a].view(full.matrix)
                        if pol == 1: 
                            line += fmt % (Asym.trace()/3*xconv3)
                        elif pol %10 == 2: 
                            out = Asym.pack().view(full.matrix)
                            out[2:4] = out[3:1:-1]
                            line += (6*fmt) % tuple(out*xconv3)

                if hyper > 0:
                    for iw in range(self.nfreqs):
                        Bsym = Bab.sum(axis=4)[iw, :, :, a].view(full.matrix)
                        if hyper == 1:
                            dip = self.Da[:,a]
                            betakk = Bsym[:,0] + Bsym[:, 3] + Bsym[:, 5]
                            line += fmt % ( 0.2 * (betakk & dip) / dip.norm2() )
                        if hyper == 2:
                            Btotsym = symmetrize_first_beta( Bsym )
                            line += 10*fmt % tuple( Btotsym )
                lines.append(line)

        return "\n".join(lines) + "\n"

    def print_atom_domain(self, n, angstrom=False):
        fmt = "%9.5f"
        if angstrom:
            xconv = 0.5291772108
        else:
            xconv = 1
        retstr = """\
---------------
Atomic domain %d
---------------
Domain center:       """  % (n+1,) + (3*fmt+"\n") % tuple(self.Rab[n, n, :]*xconv)

        print("self.max_l", self.max_l)
        if self.max_l >= 0:
            retstr += ("Nuclear charge:      " + fmt + "\n") % self.Z[n]
            retstr += ("Electronic charge:   " + fmt + "\n") % self.Qab[n, n]
            retstr += ("Total charge:        " + fmt + "\n") % (self.Z[n] + self.Qab[n,n])
        if self.max_l >= 1:
            retstr += ("Electronic dipole    " + 3*fmt + "\n") % tuple(self.Dab.sum(axis=2)[:, n])

        if self.max_l >= 2:
            retstr += ("Electronic quadrupole" + 6*fmt + "\n") % tuple((self.QUab+self.dQUab).sum(axis=2)[:,  n])

        if self.pol == 1:
            for iw, w in enumerate(self.freqs):
                retstr += ("Isotropic polarizablity (w=%g)" % w + fmt + "\n") % (
                (self.Aab + 0.5*self.dAab).sum(axis=4)[iw, :, :, n].trace()/3
                )

        if self.pol == 2:
            for iw, w in enumerate(self.freqs):
                a_lower = (self.Aab + 0.5*self.dAab).sum(axis=4)[iw, :, :, n].view(full.matrix).pack()
                retstr += ("Electronic polarizability (w=%g)" % w + 6*fmt + "\n") % tuple(
                a_lower
                )


        return retstr



# 2014.02.26 13:49:28 CET
import os
import pdb
import numpy as np
from ..qr import QR

def assert_(this, ref):
    assert np.allclose(this, ref)



def setup():
    global tmpdir
    (n, e,) = os.path.splitext(__file__)
    tmpdir = n + '.d'



def test_XXX_A2():
    XXX = QR('XDIPLEN', 'XDIPLEN', 'XDIPLEN', tmpdir=tmpdir, a2test=True)
    XXXref = 9.652e-05
    assert_(XXX, XXXref)



def test_XXX():
    XXX = QR('XDIPLEN', 'XDIPLEN', 'XDIPLEN', tmpdir=tmpdir)
    XXXref = 0.00013323
    assert_(XXX, XXXref)

def test_ZXX():
    ZXX = QR('ZDIPLEN', 'XDIPLEN', 'XDIPLEN', tmpdir=tmpdir)
    ZXXref = -1.68075251
    assert_(ZXX, ZXXref)

def test_XZX():
    XZX = QR('XDIPLEN', 'ZDIPLEN', 'XDIPLEN', tmpdir=tmpdir)
    XZXref = -1.68075251
    assert_(XZX, XZXref)

def test_XXZ():
    XXZ = QR('XDIPLEN', 'XDIPLEN', 'ZDIPLEN', tmpdir=tmpdir)
    XXZref = -1.68075251
    assert_(XXZ, XXZref)


if __name__ == '__main__':
    setup()

#+++ okay decompyling test_qr.pyc 
# decompiled 1 files: 1 okay, 0 failed, 0 verify failed
# 2014.02.26 13:49:28 CET



#!/usr/bin/env python
"""Linear response module"""
import os
from .util.timing import timing
from . import prop, sirifc, dens, rspvec, one
def LR(A, B, w=0.0, tmpdir='/tmp', absorption=False):
    """Calculate the linear response function <<A;B>> from response vector
    on RSPVEC for B, as <[kB,A]>"""
    ifcfile = os.path.join(tmpdir, 'SIRIFC')
    propfile = os.path.join(tmpdir, 'AOPROPER')
    ifc = sirifc.sirifc(ifcfile)
    a, = prop.read(A, filename=propfile, unpack=True)
    dkb =  Dk(B, freqs=(w,), ifc=ifc, tmpdir=tmpdir, absorption=absorption)
    if absorption:
        return (a&dkb[0][(B,w)], a&dkb[1][(B,w)])
    else:
        return a&dkb[(B,w)]

def Dk(*args, **kwargs):
    """Calculate the density transformed with response vector kb
    on RSPVEC for B, as d S kb - kb S d"""
    freqs = kwargs.get('freqs', (0.0,))
    ifc = kwargs.get('ifc', None)
    tmpdir = kwargs.get('tmpdir', '/tmp')
    absorption = kwargs.get('absorption', False)
    #
    # Read interface data from SIRIFC if not provided
    #
    if ifc is None: ifc = sirifc.sirifc(name=os.path.join(tmpdir, 'SIRIFC'))
    cmo = ifc.cmo
    #
    # Get densities in AO basis
    #
    dc, do = dens.ifc(os.path.join(tmpdir, "SIRIFC"), ifc)
    d = dc+do
    #
    # Get response vector (no symmetry)
    #
    kzywop = 2*ifc.nwopt

    if absorption:
        vecfile = os.path.join(tmpdir, "ABSVECS")
    else:
        vecfile = os.path.join(tmpdir, "RSPVEC")

    NB = rspvec.read(
        *args, freqs=freqs,
        propfile=vecfile
        )

    if absorption:
        ReNB = {lw: NB[lw][:len(NB[lw])//2] for lw in NB}
        ImNB = {lw: NB[lw][len(NB[lw])//2:] for lw in NB}
    #
    # Vector to matrix
    #
    cmo = ifc.cmo.unblock()
    S = one.read(filename=os.path.join(tmpdir, 'AOONEINT')).unpack().unblock()
    #(D S kb - kb S D)
    dS = d*S

    if absorption:
        Rekb = { lw:cmo*rspvec.tomat(ReNB[lw], ifc, tmpdir=tmpdir).T*cmo.T  for lw in NB }
        Imkb = { lw:cmo*rspvec.tomat(ImNB[lw], ifc, tmpdir=tmpdir).T*cmo.T  for lw in NB }
        dkb = (
            { lw: dS*Rekb[lw]-Rekb[lw]*dS.T for lw in Rekb }, 
            { lw: dS*Imkb[lw]-Imkb[lw]*dS.T for lw in Imkb }, 
            )
    else:
        kb = { lw:cmo*rspvec.tomat(NB[lw], ifc, tmpdir=tmpdir).T*cmo.T  for lw in NB }
        dkb = { lw: dS*kb[lw]-kb[lw]*dS.T for lw in kb }
    return dkb

if __name__ == "__main__":
    import sys
    from optparse import OptionParser
    op = OptionParser()
    op.add_option(
          '-t','--tmpdir',
          dest='tmpdir', default='/tmp',
          help='scratch directory [/tmp]'
          )
    try:
        opt, arg = op.parse_args(sys.argv[1:])
        Aop, Bop = arg
        tmpdir = opt.tmpdir
    except (IndexError, ValueError):
        print("Usage: %s A B" % sys.argv[0])
        sys.exit(1)

    print(LR(Aop, Bop, tmpdir))



#!/usr/bin/env python
"""Module for reading SIRIUS interface file (SIRIFC)"""
import numpy
from .util import unformatted, blocked, full

class sirifc(unformatted.FortranBinary):
    """Read data from dalton interface file"""
    ifclabel = "SIR IPH "

    def __init__(self, name="SIRIFC"):
        unformatted.FortranBinary.__init__(self, name)
        self._cmo = None
        self._dv = None
        self._pv = None
        self._fock = None
        self._fc = None
        self._fv = None

        self.find(sirifc.ifclabel)

        self.next()
        if self.reclen == 64:
           self.INT = 'q'
        elif self.reclen == 48:
           self.INT = 'i'
        elif self.reclen == 52:
           self.INT = 'i'
        else:
           raise Exception('Unconsistent first record in SIRIFC')

        self.FLOAT = 'd'

        self.potnuc, self.emy, self.eactive, self.emcscf = self.readbuf(4, self.FLOAT)
        self.istate, self.ispin, self.nactel, self.lsym  = self.readbuf(4, self.INT)

        self.next()
        self.nisht, self.nasht, self.nocct, self.norbt, self.nbast, \
        self.nconf, self.nwopt, self.nwoph, self.ncdets, self.ncmot, \
        self.nnashx, self.nnashy, self.nnorbt, self.n2orbt, self.nsym, \
        = self.readbuf(15, self.INT)

        self.muld2h = numpy.array(self.readbuf(64, self.INT)).reshape((8, 8))

        self.nrhf = numpy.array(self.readbuf(8, self.INT))
        self.nfro = numpy.array(self.readbuf(8, self.INT))
        self.nish = numpy.array(self.readbuf(8, self.INT))
        self.nash = numpy.array(self.readbuf(8, self.INT))
        self.norb = numpy.array(self.readbuf(8, self.INT))
        self.nbas = numpy.array(self.readbuf(8, self.INT))

        self.nelmn1, self.nelmx1, self.nelmn3, self.nelmx3, self.mctype \
        = self.readbuf(5, self.INT)

        self.nas1 = numpy.array(self.readbuf(8, self.INT))
        self.nas2 = numpy.array(self.readbuf(8, self.INT))
        self.nas3 = numpy.array(self.readbuf(8, self.INT))

        self.file.close()
        return


    @property
    def cmo(self):
        """Read MO coefficients"""
        if self._cmo is None:
            self.file = open(self.name,'rb')
            self.find(self.ifclabel)
            for i in range(3): self.next()
            self.file.close()
            ncmot4 = max(self.ncmot, 4)
            dbl = self.readbuf(ncmot4, self.FLOAT)
            n = 0
            self._cmo = blocked.BlockDiagonalMatrix(self.nbas, self.norb)
            for isym in range(8):
                for mo in range(self.norb[isym]):
                    for ao in range(self.nbas[isym]):
                        self._cmo.subblock[isym][ao, mo] = dbl[n]
                        n += 1
            assert(n == self.ncmot)
        return self._cmo

    @property
    def dv(self):
        """Get active density matrix"""
        if self._dv is None:
            self.file = open(self.name,'rb')
            self.find(self.ifclabel)
            for i in range(5): 
                self.next()
            mmashx = max(self.nnashx, 4)
            dbl = self.readbuf(self.nnashx, self.FLOAT)
            self._dv = full.triangular.init(dbl)
        return self._dv

    @property
    def pv(self):
        """Get two-electron density"""
        if self._pv is None:
            self.file = open(self.name, 'rb')
            self.find(self.ifclabel)
            for i in range(7): self.next()
            self.file.close()
            m2ashy = max(self.nnashx**2, 4)
            dbl = self.readbuf(m2ashy, self.FLOAT)
            self._pv = full.matrix((self.nnashx, self.nnashx))
            n = 0
            for i in range(self.nnashx):
                for j in range(self.nnashx):
                    self._pv[j, i] = dbl[n]
                    n += 1
            assert(n == self.nnashx**2)
        return self._pv

    @property
    def fock(self):
        """Read Fock matrix (MO)"""
        if self._fock is None:
            self.file = open(self.name, 'rb')
            self.find(self.ifclabel)
            for i in range(6): self.next()
            self.file.close()
            m2orbt = max(self.n2orbt, 4)
            dbl = self.readbuf(m2orbt, self.FLOAT)
            self._fock = blocked.matrix(self.norb, self.norb)
            n = 0
            for isym in range(8):
                for i in range(self.norb[isym]):
                    for j in range(self.norb[isym]):
                        self._fock.subblock[isym][j, i] = dbl[n]
                        n += 1
            assert (n == self.n2orbt)
        return self._fock

    @property
    def fc(self):
        """Read inactive Fock matrix (MO)"""
        if self._fc is None:
            self.file = open(self.name, 'rb')
            self.find(self.ifclabel)
            for i in range(8): self.next()
            mmorbt = max(self.nnorbt, 4)
            dbl = self.readbuf(mmorbt, self.FLOAT)
            self._fc = blocked.triangular(self.norb)
            n = 0
            for isym in range(8):
                ij = 0
                for i in range(self.norb[isym]):
                    for j in range(i+1):
                        self._fc.subblock[isym][i, j] = dbl[ij]
                        ij += 1
                n += ij
            assert (n == self.nnorbt)
        return self._fc

    @property
    def fv(self):
        """Get active Fock matrix"""
        if self._fv is None:
            self.file = open(self.name, 'rb')
            self.find(self.ifclabel)
            for i in range(9): self.next()
            mmorbt = max(self.nnorbt, 4)
            dbl = self.readbuf(mmorbt, self.FLOAT)
            self._fv = blocked.triangular(self.norb)
            n = 0
            for isym in range(8):
                ij = 0
                for i in range(self.norb[isym]):
                    for j in range(i+1):
                        self._fv.subblock[isym][i, j] = dbl[ij]
                        ij += 1
                n += ij
            assert (n == self.nnorbt)
        return self._fv

    #dv = property(fget=get_dv)
    #pv = property(fget=get_pv)
    #cmo = property(fget=get_cmo)
    #fock = property(fget=get_fock)
    #fc = property(fget=get_fc)
    #fv = property(fget=get_fv)

    def __repr__(self):
        retstr = ""
        retstr += "Nuclear Potential Energy: %12.6f\n" % self.potnuc
        retstr += "Electronic energy       : %12.6f\n" % self.emy
        retstr += "Active energy           : %12.6f\n" % self.eactive
        retstr += "MCSCF energy            : %12.6f\n" % self.emcscf
        retstr += "State                   : %d\n" % self.istate
        retstr += "Spin                    : %d\n" % self.ispin
        retstr += "Active electrons        : %d\n" % self.nactel
        retstr += "Symmetry                : %d\n" % self.lsym
        retstr += "NISHT                   : %d\n" % self.nisht
        retstr += "NASHT                   : %d\n" % self.nasht
        retstr += "NOCCT                   : %d\n" % self.nocct
        retstr += "NORBT                   : %d\n" % self.norbt
        retstr += "NBAST                   : %d\n" % self.nbast
        retstr += "NCONF                   : %d\n" % self.nconf
        retstr += "NWOPT                   : %d\n" % self.nwopt
        retstr += "NWOPH                   : %d\n" % self.nwoph
        retstr += "NCDETS                  : %d\n" % self.ncdets
        retstr += "NCMOT                   : %d\n" % self.ncmot
        retstr += "NNASHX                  : %d\n" % self.nnashx
        retstr += "NNASHY                  : %d\n" % self.nnashy
        retstr += "NNORBT                  : %d\n" % self.nnorbt
        retstr += "N2ORBT                  : %d\n" % self.n2orbt
        retstr += "NSYM                    : %d\n" % self.nsym
        retstr += "MULD2H:\n"
        for i in range(8):
            retstr += "   "
            for j in range(8):
                retstr += " %d" % self.muld2h[i, j]
            retstr += "\n"
       
        def strvec(lab, vec):
            """ String representation of vector"""
            locstr = lab + ":"
            for i in range(8): locstr += " %d" % vec[i]
            return locstr + "\n"

        retstr += strvec("NRHF", self.nrhf)
        retstr += strvec("NFRO", self.nfro)
        retstr += strvec("NISH", self.nish)
        retstr += strvec("NASH", self.nash)
        retstr += strvec("NORB", self.norb)
        retstr += strvec("NBAS", self.nbas)
        retstr += "NELMN1                  : %d\n" % self.nelmn1
        retstr += "NELMX1                  : %d\n" % self.nelmx1
        retstr += "NELMN3                  : %d\n" % self.nelmn3
        retstr += "NELMX3                  : %d\n" % self.nelmx3
        retstr += "MCTYPE                  : %d\n" % self.mctype
        retstr += strvec("NAS1", self.nas1)
        retstr += strvec("NAS2", self.nas2)
        retstr += strvec("NAS3", self.nas3)
        retstr += "CMO" + str(self.cmo)
        retstr += "DV\n" + str(self.dv) + "\n"
        retstr += "FOCK" + str(self.fock)
        retstr += "PV\n" + str(self.pv) + "\n"
        retstr += "FC" + str(self.fc)
        retstr += "FV" + str(self.fv)
        return retstr

if __name__ == "__main__":
    import sys
    try:
        filename = sys.argv[1]
    except IndexError as e:
        print("IndexError", e)
        print("Usage: %s [path]/SIRIFC" % sys.argv[0])
        sys.exit(1)

    ifc = sirifc(name=filename)
    print(ifc)
        



import unittest
import os
import numpy as np
from ..rspvec import read
from ..lr import LR

class TestALR(unittest.TestCase):

    def setUp(self):
        n, _ = os.path.splitext(__file__)
        self.tmpdir = n + ".d"

    def tmp(self, file_):
        return os.path.join(self.tmpdir, file_)

    def test_first_xvec(self):
        vecs = read('XDIPLEN', freqs=(0.4425,), propfile=self.tmp('ABSVECS'))
        np.testing.assert_allclose(
            vecs[('XDIPLEN', 0.4425, 0.0)][:10],
            [
           -1.7563977677089527e-09,    0.0015755554234863397,
           -5.504742822192291e-10,     0.002779217059896619,
           -0.0034886672773900523,    7.051844231481651e-10,
           -0.005062366558336546,    -0.006285259584578197,
           -0.008674385885382397,   -7.986196390011771e-11
            ])

    def test_first_yvec(self):
        vecs = read('XDIPLEN', 'YDIPLEN', freqs=(0.4425,), propfile=self.tmp('ABSVECS'))
        np.testing.assert_allclose(
            vecs[('YDIPLEN', 0.4425, 0.0)][:10],
            [
          -3.062472422126098e-08,    0.0027792414636958196,
          -1.1898942484099601e-07,   -0.0015755667065659468,
          -0.005062426714257009,  1.0479887117851179e-07,
           0.00348869833460343,    -0.008674366708527649,
           0.006285236842617945,    2.145201913436897e-07
            ])

    def test_number_vecs(self):
        vecs = read(
            'XDIPLEN', 'YDIPLEN', 'ZDIPLEN', freqs=(0.4425, 0.49), propfile=self.tmp('ABSVECS'),
            lr_vecs=True
            )
        self.assertEqual(len(vecs), 6)

    def test_real_alpha_xx(self):
        w = 0.4425
        re_alpha, _ = LR('XDIPLEN', 'XDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, 30.854533, delta=1e-6)

    def test_real_alpha_yy(self):
        w = 0.4450
        re_alpha, _ = LR('YDIPLEN', 'YDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, 31.544221, delta=1e-6)

    def test_real_alpha_zz(self):
        w = 0.4475
        re_alpha, _ = LR('ZDIPLEN', 'ZDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, 32.275296, delta=1e-6)

    def test_real_alpha_xy(self):
        w = 0.4425
        re_alpha, _ = LR('XDIPLEN', 'YDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, -0.000004, delta=1e-6)

    def test_real_alpha_xz(self):
        w = 0.4425
        re_alpha, _ = LR('XDIPLEN', 'ZDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, 0.000000, delta=1e-6)


    def test_real_alpha_yz(self):
        w = 0.4425
        re_alpha, _ = LR('XDIPLEN', 'ZDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-re_alpha, 0.000000, delta=1e-6)


    def test_real_im(self):
        w = 0.4425
        _, im_alpha = LR('XDIPLEN', 'XDIPLEN', w, tmpdir=self.tmpdir, absorption=True)
        self.assertAlmostEqual(-im_alpha, 1.228334, delta=1e-6)

            
if __name__ == "__main__":
    unittest.main()



import unittest
import os
import numpy as np
from ..lr import LR

class TestLR(unittest.TestCase):

    def setUp(self):
        n, _ = os.path.splitext(__file__)
        self.tmpdir = n + ".d"

    def test_XX(self):
        XX = LR('XDIPLEN', 'XDIPLEN', 0, self.tmpdir)
        XXref = -2.461169664950
        self.assertAlmostEqual(XX, XXref)

    def test_YY(self):
        YY = LR('YDIPLEN', 'YDIPLEN', 0, self.tmpdir)
        YYref = -6.184500121159
        self.assertAlmostEqual(YY, YYref)

    def test_YZ(self):
        YZ = LR('YDIPLEN', 'ZDIPLEN', 0, self.tmpdir)
        YZref = 0.016158638023
        self.assertAlmostEqual(YZ, YZref)

    def test_ZZ(self):
        ZZ = LR('ZDIPLEN', 'ZDIPLEN', 0, self.tmpdir)
        ZZref = -10.308181624834
        self.assertAlmostEqual(ZZ, ZZref)

if __name__ == "__main__":
    unittest.main()



import unittest
import numpy
from ..subblocked import matrix as SubBlockedMatrix


class NewTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_str(self):
        asb = SubBlockedMatrix([2,1], [2,1])
        print(asb)
        self.assertEqual(str(asb), """
Block (1,1)

 (2, 2) 
              Column   1    Column   2

Block (1,2)

 (2, 1) 
              Column   1

Block (2,1)

 (1, 2) 
              Column   1    Column   2

Block (2,2)

 (1, 1) 
              Column   1
"""
        )


    def test_tranpose_diagonal_block(self):
        A = SubBlockedMatrix([2], [2])
        A.subblock[0][0][:, :] = [[1., 2.], [3., 4]]
        AT = A.T()
        numpy.testing.assert_allclose(AT.subblock[0][0], [[1., 3.], [2., 4.]])

    def test_tranpose_offdiagonal_block(self):
        A = SubBlockedMatrix([1, 2], [1, 2])
        A.subblock[0][1][:, :] = [[1., 2.]]
        AT = A.T()
        numpy.testing.assert_allclose(AT.subblock[1][0], [[1.], [2.]])

    def test_mul(self):
        A = SubBlockedMatrix([2, 1], [2, 1])
        A.subblock[0][0][:, :] = [[1., 0.], [0., 1.]]
        A.subblock[1][1][:, :] = 1.0

        A2 = A*A
        numpy.testing.assert_allclose(A2.subblock[0][0], A2.subblock[0][0])

    def test_random(self):
        A = SubBlockedMatrix([2], [2])
        numpy.random.seed(0)
        A.random()
        numpy.testing.assert_allclose(A.subblock[0][0], [
            [ 0.5488135 ,  0.71518937],
            [ 0.60276338,  0.54488318]
           ])

    def test_unblock(self):
        A = SubBlockedMatrix([2], [2])
        numpy.random.seed(0)
        A.random()
        Afull = A.unblock()
        numpy.testing.assert_allclose(Afull, [
            [ 0.5488135 ,  0.71518937],
            [ 0.60276338,  0.54488318]
           ])


if __name__ == "__main__": #pragma: no cover
    unittest.main()



import os 
import numpy as np
from ..daltools.util import full

import re
thisdir  = os.path.dirname(__file__)
case = "h2o_beta"
tmpdir=os.path.join(thisdir, case, 'tmp')
exec('from . import %s_data as ref'%case)

from ..core import penalty_function, xtang, pairs, MolFrag

def assert_(this, ref, atol=1e-5, text=None):
    if text: print(text)
    print(this, ref)
    print("Max deviation", np.amax(this - ref))
    assert np.allclose(this, ref, atol=atol)

def assert_str(this, ref, text=None):
    if text: print(text)
    print(this, ref)
    # allow string inequality from round-off errors
    assert this.replace("-0.000", " 0.000") == ref.replace("-0.000", " 0.000")


def setup():
    global m
    global ff
# modify Gagliardi penalty function to include unit conversion bug
    m = MolFrag(tmpdir, freqs=(0.0,), pf=penalty_function(2.0/xtang**2))
    ff = ref.ff

def test_nuclear_charge():
    Z = m.Z
    assert_(Z, ref.Z)

def test_coordinates_au():
    R = m.R
    assert_(R, ref.R)

def test_default_gauge():
    assert_(ref.Rc, m.Rc)

def test_total_charge():
    Qtot = m.Qab.sum()
    assert_(Qtot, ref.Qtot)

def test_charge():
    Qaa = m.Qab.diagonal()
    assert_(ref.Q, Qaa)

def test_total_dipole():
    # molecular dipole moment wrt gauge center gc
    Dtot = m.Dab.sum(axis=2).sum(axis=1).view(full.matrix)
    Qa = m.Qab.diagonal()
    Q = Qa.sum()
    Dtot += Qa*m.R - Q*m.Rc
    assert_(Dtot, ref.Dtot)

def test_dipole_allbonds():
    D = full.matrix(ref.D.shape)
    Dab = m.Dab
    for ab, a, b in pairs(m.noa):
        D[:, ab] += Dab[:, a, b ] 
        if a != b: D[:, ab] += Dab[:, b, a] 
    assert_(D, ref.D)

def test_dipole_allbonds_sym():
    Dsym = m.Dsym
    assert_(Dsym, ref.D)

def test_dipole_nobonds():
    Daa = m.Dab.sum(axis=2).view(full.matrix)
    assert_(Daa, ref.Daa)

def test_quadrupole_total():
    QUc = m.QUc
    assert_(QUc, ref.QUc)
    
def test_nuclear_quadrupole():
    QUN = m.QUN
    assert_(QUN, ref.QUN)

def test_quadrupole_allbonds():
    QU = full.matrix(ref.QU.shape)
    QUab = m.QUab
    for ab, a, b in pairs(m.noa):
        QU[:, ab] += QUab[:, a, b ] 
        if a != b: QU[:, ab] += QUab[:, b, a] 
    assert_(QU, ref.QU)

def test_quadrupole_allbonds_sym():
    QUsym = m.QUsym
    assert_(QUsym, ref.QU)

def test_quadrupole_nobonds():

    QUaa = (m.QUab + m.dQUab).sum(axis=2).view(full.matrix)
    assert_(QUaa, ref.QUaa)


def test_Fab():

    Fab = m.Fab
    assert_(Fab, ref.Fab)

def test_molcas_shift():
    Fab = m.Fab
    Lab = Fab + m.sf(Fab)

    assert_(Lab, ref.Lab)

def test_total_charge_shift():
    dQ = m.dQa[0].sum(axis=0).view(full.matrix)
    dQref = [0., 0., 0.]
    assert_(dQref, dQ)

def test_total_charge_shift2():
    d2Q = m.d2Qa[0].sum(axis=0).view(full.matrix)
    d2Qref = [0., 0., 0., 0., 0., 0.]
    assert_(d2Qref, d2Q)

def test_atomic_charge_shift():
    dQa = m.dQa[0]
    dQaref = (ref.dQa[:, 1::2] - ref.dQa[:, 2::2])/(2*ff)
    assert_(dQa, dQaref, .006)

def test_lagrangian():
# values per "perturbation" as in atomic_charge_shift below
    la = m.la[0]
    laref = (ref.la[:,0:6:2] - ref.la[:,1:6:2])/(2*ff)
# The sign difference is because mocas sets up rhs with opposite sign
    assert_(-laref, la, atol=100)

def test_bond_charge_shift():
    dQab = m.dQab[0]
    noa = m.noa


    dQabref = (ref.dQab[:, 1:7:2] - ref.dQab[:, 2:7:2])/(2*ff)
    dQabcmp = full.matrix((3, 3))
    ab = 0
    for a in range(noa):
        for b in range(a):
            dQabcmp[ab, :] = dQab[a, b, :]
            ab += 1
# The sign difference is because mocas sets up rhs with opposite sign
    assert_(-dQabref, dQabcmp, 0.006)

def test_bond_charge_shift_sum():
    dQa  = m.dQab[0].sum(axis=1).view(full.matrix)
    dQaref = m.dQa[0]
    assert_(dQa, dQaref)

def test_bond_charge_shift_sum2():
    d2Qa  = m.d2Qab[0].sum(axis=1).view(full.matrix)
    d2Qaref = m.d2Qa[0]
    assert_(d2Qa, d2Qaref)


def test_polarizability_total():

    Am = m.Am[0]

    assert_(Am, ref.Am, 0.015)

def test_beta_zxx():
    r = m.x
    D2k = m.D2k
    z = r[2].unblock()
    xx = D2k[('XDIPLEN XDIPLEN ', 0.0, 0.0)].unblock()
    bzxx = - z&xx
    assert_(bzxx, ref.Bm[2, 0], .005)

def test_beta_xzx():
    r = m.x
    D2k = m.D2k
    x = r[0].unblock()
    zx = D2k[('ZDIPLEN XDIPLEN ', 0.0, 0.0)].unblock()
    bxzx = - x&zx
    assert_(bxzx, ref.Bm[0, 2], .005)

def test_beta_xxz():
    r = m.x
    D2k = m.D2k
    x = r[0].unblock()
    xz = D2k[('XDIPLEN ZDIPLEN ', 0.0, 0.0)].unblock()
    bxxz = - x&xz
    assert_(bxxz, ref.Bm[0, 2], .005)

def test_beta_yyz():
    r = m.x
    D2k = m.D2k
    y = r[1].unblock()
    yz = D2k[('YDIPLEN ZDIPLEN ', 0.0, 0.0)].unblock()
    byyz = - y&yz
    assert_(byyz, ref.Bm[1, 4], .005)

def test_beta_zyy():
    r = m.x
    D2k = m.D2k
    z = r[2].unblock()
    yy = D2k[('YDIPLEN YDIPLEN ', 0.0, 0.0)].unblock()
    bzyy = - z&yy
    assert_(bzyy, ref.Bm[2, 3], .005)

def test_beta_zzz():
    r = m.x
    D2k = m.D2k
    z = r[2].unblock()
    zz = D2k[('ZDIPLEN ZDIPLEN ', 0.0, 0.0)].unblock()
    bzzz = - z&zz
    assert_(bzzz, ref.Bm[2, 5], .005)

def test_hyperpolarizability_total():
    Bm = m.Bm[0]
    Bab = m.Bab.sum(axis=4).sum(axis=3)
    ref.Bm
    assert_(Bm, ref.Bm, .005)

def notest_dynamic_polarizability_total():

    Amw = m.Am[1]

    assert_(Amw, ref.Amw, 0.015)
        
def test_polarizability_allbonds_molcas_internal():

    O = ref.O
    H1O = ref.H1O
    H1 = ref.H1
    H2O = ref.H2O
    H2H1 = ref.H2H1
    H2 = ref.H2
    rMP = ref.rMP

    RO, RH1, RH2 = m.R
    ROx, ROy, ROz = RO
    RH1x, RH1y, RH1z = RH1
    RH2x, RH2y, RH2z = RH2
    

    ihff = 1/(2*ff)

    q, x, y, z = range(4)
    dx1, dx2, dy1, dy2, dz1, dz2 = 1, 2, 3, 4, 5, 6
    o, h1o, h1, h2o, h2h1, h2 = range(6)

    Oxx = ihff*(rMP[x, dx1, o] - rMP[x, dx2, o])
    Oyx = ihff*(rMP[y, dx1, o] - rMP[y, dx2, o]
        +       rMP[x, dy1, o] - rMP[x, dy2, o])/2
    Oyy = ihff*(rMP[y, dy1, o] - rMP[y, dy2, o])
    Ozx = ihff*(rMP[z, dx1, o] - rMP[z, dx2, o]
        +       rMP[x, dz1, o] - rMP[x, dz2, o])/2
    Ozy = ihff*(rMP[z, dy1, o] - rMP[z, dy2, o]
        +       rMP[y, dz1, o] - rMP[y, dz2, o])/2
    Ozz = ihff*(rMP[z, dz1, o] - rMP[z, dz2, o])
    H1Oxx = ihff*(rMP[x, dx1, h1o] - rMP[x, dx2, h1o] \
          - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1x-ROx))
    H1Oyx = ihff*(
         (rMP[y, dx1, h1o] - rMP[y, dx2, h1o] 
        + rMP[x, dy1, h1o] - rMP[x, dy2, h1o])/2
       - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1y-ROy)
#      - (rMP[0, dy1, h1o] - rMP[0, dy2, h1o])*(RH1x-ROx) THIS IS REALLY... A BUG?
       )
    H1Oyy = ihff*(rMP[y, dy1, h1o] - rMP[y, dy2, h1o] - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH1y-ROy))
    H1Ozx = ihff*(
        (rMP[z, dx1, h1o] - rMP[z, dx2, h1o]
       + rMP[x, dz1, h1o] - rMP[x, dz2, h1o])/2
      - (rMP[q, dx1, h1o] - rMP[q, dx2, h1o])*(RH1z-ROz)
#             - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1x-ROx) #THIS IS REALLY... A BUG?
            )
    H1Ozy = ihff*(
        (rMP[z, dy1, h1o] - rMP[z, dy2, h1o]
       + rMP[y, dz1, h1o] - rMP[y, dz2, h1o])/2
      - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH1z-ROz)
#     - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1y-ROy) THIS IS REALLY... A BUG?
        )
    H1Ozz = ihff*(rMP[z, dz1, h1o] - rMP[z, dz2, h1o] - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH1z-ROz))
    H1xx = ihff*(rMP[x, dx1, h1] - rMP[x, dx2, h1])
    H1yx = (ihff*(rMP[y, dx1, h1] - rMP[y, dx2, h1])
         +  ihff*(rMP[x, dy1, h1] - rMP[x, dy2, h1]))/2
    H1yy = ihff*(rMP[y, dy1, h1] - rMP[y, dy2, h1])
    H1zx = (ihff*(rMP[z, dx1, h1] - rMP[z, dx2, h1])
         +  ihff*(rMP[x, dz1, h1] - rMP[x, dz2, h1]))/2
    H1zy = (ihff*(rMP[z, dy1, h1] - rMP[z, dy2, h1])
         +  ihff*(rMP[y, dz1, h1] - rMP[y, dz2, h1]))/2
    H1zz = ihff*(rMP[z, dz1, h1] - rMP[z, dz2, h1])
    H2Oxx = ihff*(rMP[x, dx1, h2o] - rMP[x, dx2, h2o] - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2x-ROx))
    H2Oyx = ihff*(
        (rMP[y, dx1, h2o] - rMP[y, dx2, h2o] 
       + rMP[x, dy1, h2o] - rMP[x, dy2, h2o])/2
       - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2y-ROy)
#      - (rMP[q, dy1, h1o] - rMP[q, dy2, h1o])*(RH2x-ROx) THIS IS REALLY... A BUG?
       )
    H2Oyy = ihff*(rMP[y, dy1, h2o] - rMP[y, dy2, h2o] - (rMP[q, dy1, h2o] - rMP[q, dy2, h2o])*(RH2y-ROy))
    H2Ozx = ihff*(
        (rMP[z, dx1, h2o] - rMP[z, dx2, h2o]
       + rMP[x, dz1, h2o] - rMP[x, dz2, h2o])/2
      - (rMP[q, dx1, h2o] - rMP[q, dx2, h2o])*(RH2z-ROz)
#             - (rMP[q, dz1, h1o] - rMP[q, dz2, h1o])*(RH2x-ROx) #THIS IS REALLY... A BUG?
            )
    H2Ozy = ihff*(
        (rMP[z, dy1, h2o] - rMP[z, dy2, h2o]
       + rMP[y, dz1, h2o] - rMP[y, dz2, h2o])/2
      - (rMP[q, dy1, h2o] - rMP[q, dy2, h2o])*(RH2z-ROz)
#     - (rMP[q, dz1, h2o] - rMP[q, dz2, h2o])*(RH2y-ROy) THIS IS REALLY... A BUG?
        )
    H2Ozz = ihff*(rMP[z, dz1, h2o] - rMP[z, dz2, h2o] - (rMP[q, dz1, h2o] - rMP[q, dz2, h2o])*(RH2z-ROz))
    H2H1xx = ihff*(rMP[x, dx1, h2h1] - rMP[x, dx2, h2h1] - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH2x-RH1x))
    H2H1yx = ihff*(
        (rMP[y, dx1, h2h1] - rMP[y, dx2, h2h1] 
       + rMP[x, dy1, h2h1] - rMP[x, dy2, h2h1])/2
       - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH1y-ROy)
#      - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH1x-ROx) THIS IS REALLY... A BUG?
       )
    H2H1yy = ihff*(rMP[y, dy1, h2h1] - rMP[y, dy2, h2h1] - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH2y-RH1y))
    H2H1zx = ihff*(
        (rMP[z, dx1, h2h1] - rMP[z, dx2, h2h1]
       + rMP[x, dz1, h2h1] - rMP[x, dz2, h2h1])/2
      - (rMP[q, dx1, h2h1] - rMP[q, dx2, h2h1])*(RH1z-ROz)
#     - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH1x-ROx) #THIS IS REALLY... A BUG?
            )
    H2H1zy = ihff*(
        (rMP[z, dy1, h2h1] - rMP[z, dy2, h2h1]
       + rMP[y, dz1, h2h1] - rMP[y, dz2, h2h1])/2
      - (rMP[q, dy1, h2h1] - rMP[q, dy2, h2h1])*(RH1z-ROz)
#     - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH1y-RO[1]) THIS IS REALLY... A BUG?
        )
    H2H1zz = ihff*(rMP[z, dz1, h2h1] - rMP[z, dz2, h2h1] - (rMP[q, dz1, h2h1] - rMP[q, dz2, h2h1])*(RH2z-RH1z))
    H2xx = ihff*(rMP[x, dx1, h2] - rMP[x, dx2, h2])
    H2yx = (ihff*(rMP[y, dx1, h2] - rMP[y, dx2, h2])
         +  ihff*(rMP[x, dy1, h2] - rMP[x, dy2, h2]))/2
    H2yy = ihff*(rMP[y, dy1, h2] - rMP[y, dy2, h2])
    H2zx = (ihff*(rMP[z, dx1, h2] - rMP[z, dx2, h2])
         +  ihff*(rMP[x, dz1, h2] - rMP[x, dz2, h2]))/2
    H2zy = (ihff*(rMP[z, dy1, h2] - rMP[z, dy2, h2])
         +  ihff*(rMP[y, dz1, h2] - rMP[y, dz2, h2]))/2
    H2zz = ihff*(rMP[z, dz1, h2] - rMP[z, dz2, h2])

    comp = ("XX", "yx", "yy", "zx", "zy", "zz")
    bond = ("O", "H1O", "H1", "H2O", "H2H1", "H2")

    assert_(O[0], Oxx, text="Oxx")
    assert_(O[1], Oyx, text="Oyx")
    assert_(O[2], Oyy, text="Oyy")
    assert_(O[3], Ozx, text="Ozx")
    assert_(O[4], Ozy, text="Ozy")
    assert_(O[5], Ozz, text="Ozz")
    assert_(H1O[0], H1Oxx, text="H1Oxx")
    assert_(H1O[1], H1Oyx, text="H1Oyx")
    assert_(H1O[2], H1Oyy, text="H1Oyy")
    assert_(H1O[3], H1Ozx, text="H1Ozx")
    assert_(H1O[4], H1Ozy, text="H1Ozy")
    assert_(H1O[5], H1Ozz, text="H1Ozz")
    assert_(H1[0], H1xx, text="H1xx")
    assert_(H1[1], H1yx, text="H1yx")
    assert_(H1[2], H1yy, text="H1yy")
    assert_(H1[3], H1zx, text="H1zx")
    assert_(H1[4], H1zy, text="H1zy")
    assert_(H1[5], H1zz, text="H1zz")
    assert_(H2O[0], H2Oxx, text="H2Oxx")
    assert_(H2O[1], H2Oyx, text="H2Oyx")
    assert_(H2O[2], H2Oyy, text="H2Oyy")
    assert_(H2O[3], H2Ozx, text="H2Ozx")
    assert_(H2O[4], H2Ozy, text="H2Ozy")
    assert_(H2O[5], H2Ozz, text="H2Ozz")
    assert_(H2H1[0], H2H1xx, text="H2H1xx")
    assert_(H2H1[1], H2H1yx, text="H2H1yx")
    assert_(H2H1[2], H2H1yy, text="H2H1yy")
    assert_(H2H1[3], H2H1zx, text="H2H1zx")
    assert_(H2H1[4], H2H1zy, text="H2H1zy")
    assert_(H2H1[5], H2H1zz, text="H2H1zz")
    assert_(H2[0], H2xx, text="H2xx")
    assert_(H2[1], H2yx, text="H2yx")
    assert_(H2[2], H2yy, text="H2yy")
    assert_(H2[3], H2zx, text="H2zx")
    assert_(H2[4], H2zy, text="H2zy")
    assert_(H2[5], H2zz, text="H2zz")

def test_altint():
    R = m.R
    rMP = ref.rMP
    diff = [(1, 2), (3, 4), (5, 6)]
    atoms = (0, 2, 5) 
    bonds = (1, 3, 4)
    ablab = ("O", "H1O", "H1", "H2O", "H2H1", "H2")
    ijlab = ("xx", "yx", "yy", "zx", "zy", "zz")

    pol = np.zeros((6, m.noa*(m.noa+1)//2))
    for ab, a, b in pairs(m.noa):
        for ij, i, j in pairs(3):
            #from pdb import set_trace; set_trace()
            i1, i2 = diff[i]
            j1, j2 = diff[j]
            pol[ij, ab] += (rMP[i+1, j1, ab] - rMP[i+1, j2, ab]
                        +   rMP[j+1, i1, ab] - rMP[j+1, i2, ab])/(4*ff)
            if ab in bonds:
                pol[ij, ab] -= (R[a][i]-R[b][i])*(rMP[0, j1, ab] - rMP[0, j2, ab])/(2*ff)
            assert_(ref.Aab[ij, ab], pol[ij, ab], text="%s%s"%(ablab[ab], ijlab[ij]))

def test_polarizability_allbonds_atoms():

    Aab = m.Aab[0] #+ m.dAab
    noa = m.noa

    Acmp=full.matrix(ref.Aab.shape)
    
    ab = 0
    for a in range(noa):
        for b in range(a):
            Acmp[:, ab] = (Aab[:, :, a, b] + Aab[:, :, b, a]).pack()
            ab += 1
        Acmp[:, ab] = Aab[:, :, a, a].pack()
        ab += 1
    # atoms
    assert_(ref.Aab[:, 0], Acmp[:, 0], .005)
    assert_(ref.Aab[:, 2], Acmp[:, 2], .005)
    assert_(ref.Aab[:, 5], Acmp[:, 5], .005)

def test_polarizability_allbonds_bonds():

    Aab = m.Aab[0] + m.dAab[0]/2
    noa = m.noa

    Acmp=full.matrix(ref.Aab.shape)
    
    ab = 0
    for a in range(noa):
        for b in range(a):
            Acmp[:, ab] = (Aab[:, :, a, b] + Aab[:, :, b, a]).pack()
            ab += 1
        Acmp[:, ab] = Aab[:, :, a, a].pack()
        ab += 1
    # atoms
    assert_(ref.Aab[:, 1], Acmp[:, 1], .150, 'H1O')
    assert_(ref.Aab[:, 3], Acmp[:, 3], .150, 'H2O')
    assert_(ref.Aab[:, 4], Acmp[:, 4], .005, 'H2H1')
    

def test_polarizability_nobonds():

    Aab = m.Aab[0] + m.dAab[0]/2
    noa = m.noa

    Acmp = full.matrix((6, noa ))
    Aa = Aab.sum(axis=3).view(full.matrix)
    
    ab = 0
    for a in range(noa):
        Acmp[:, a,] = Aa[:, :, a].pack()

    # atoms
    assert_(Acmp, ref.Aa, 0.07)

def test_potfile_PAn0():
    PAn0 = m.output_potential_file(maxl=-1, pol=0, hyper=0)
    assert_str(PAn0, ref.PAn0)

def test_potfile_PA00():
    PA00 = m.output_potential_file(maxl=0, pol=0, hyper=0)
    assert_str(PA00, ref.PA00)

def test_potfile_PA10():
    PA10 = m.output_potential_file(maxl=1, pol=0, hyper=0)
    assert_str(PA10, ref.PA10)

def test_potfile_PA20():
    PA20 = m.output_potential_file(maxl=2, pol=0, hyper=0)
    assert_str(PA20, ref.PA20)

def test_potfile_PA21():
    PA21 = m.output_potential_file(maxl=2, pol=1, hyper=0)
    assert_str(PA21, ref.PA21)

def test_potfile_PA22():
    PA22 = m.output_potential_file(maxl=2, pol=2, hyper=0)
    assert_str(PA22, ref.PA22)

if __name__ == "__main__":
    setup()
    test_hyperpolarizability_total()



import unittest
import os
import numpy as np
from .. import one
from ..util import blocked, full

class TestOne(unittest.TestCase):

    def setUp(self):
        n, _ = os.path.splitext(__file__)
        self.tmpdir = n + ".d"
        self.aooneint = os.path.join(self.tmpdir, 'AOONEINT')
        self.header = one.readhead(self.aooneint)

    def test_header_title(self):
        self.assertIn(b"CH2O", self.header["ttitle"])

    def test_header_naos(self):
        self.assertTupleEqual(self.header["naos"], (7, 2, 3, 0))

    def test_header_nsym(self):
        self.assertEqual(self.header["nsym"], 4)

    def test_header_potnuc(self):
        self.assertAlmostEqual(self.header["potnuc"], 33.941796143849)

    def test_isordk_nucdep(self):
        isordk = one.readisordk(self.aooneint)
        self.assertEqual(isordk["nucdep"], 4)
    
    def test_isordk_chrn(self):
        isordk = one.readisordk(self.aooneint)
        self.assertTupleEqual(isordk["chrn"][:3], (6., 8., 1.))

    #@unittest.skip('Wrong format here')
    def test_isordk_cooo(self):
        isordk = one.readisordk(self.aooneint)
        C = [0.0000000000, 0.0000000000,-0.8802560791]
        O = [0.0000000000, 0.0000000000, 1.0094700459]
        H1 = [0.0000000000, 1.8897261250, -2.7699822041]
        H2 = [0.0000000000, -1.8897261250, -2.7699822041]
        np.testing.assert_almost_equal(isordk["cooo"][0::500], C)
        np.testing.assert_almost_equal(isordk["cooo"][1::500], O)
        np.testing.assert_almost_equal(isordk["cooo"][2::500], H1)
        np.testing.assert_almost_equal(isordk["cooo"][3::500], H2)

    def test_scfinp(self):
        scfinp = one.readscfinp(self.aooneint)
        self.assertEqual(scfinp["nsym"], 4)
        coor_bohr = (
            0.0000000000, 0.0000000000,-0.8802560791,
            0.0000000000, 0.0000000000, 1.0094700459,
            0.0000000000, 1.8897261250, -2.7699822041,
            0.0000000000, -1.8897261250, -2.7699822041,
            )

        np.testing.assert_almost_equal(scfinp["cooo"], coor_bohr)

    def test_overlap(self):
        Sref = blocked.triangular.init([
            [ 1.00000000,
            0.24836239, 1.00000000,
            0.00000000, 0.00000000, 1.00000000,
            0.00005599, 0.05942391, 0.10228506, 1.00000000,
            0.07053810, 0.49046481, 0.53543774, 0.23670394, 1.00000000,
           -0.11599611,-0.37963762,-0.27726502,-0.00000000, 0.00000000, 1.00000000,
            0.05926125, 0.64602901, -0.48763174,0.00548289, 0.09221295,-0.09620875, 2.23843306],
            [1.00000000, 
             0.31956952, 1.00000000],
            [1.00000000,
             0.31956952, 1.00000000,
             0.48763174, 0.04810438, 1.76156694]
        ])


        S = one.read("OVERLAP", self.aooneint)
        np.testing.assert_almost_equal(S.subblock[0], Sref.subblock[0])
        np.testing.assert_almost_equal(S.subblock[1], Sref.subblock[1])
        np.testing.assert_almost_equal(S.subblock[2], Sref.subblock[2])

if __name__ == "__main__":
    unittest.main()



import unittest
try:
    import mock
except ImportError:
    from unittest import mock
import os
import sys
from ..unformatted import FortranBinary
import numpy as np

def mock_it(builtin_name):
   name = ('builtins.%s' if sys.version_info >= (3,) else '__builtin__.%s') % builtin_name
   return mock.patch(name)

class TestFortranBinary(unittest.TestCase):

    def setUp(self):
        self.tdir, _ = os.path.split(__file__)

    def test_1(self):
        """Read int, float

          integer, parameter :: n = 3
          double precision x(n)
          x = (/ 1.0D0, 2.0D0, 3.0D0 /)
          open(1, file='fort.1', status='new', form='unformatted')
          write(1) n
          write(1) x
          close(1)
          end
        """
        ffile = os.path.join(self.tdir, 'fort.1')
        fb = FortranBinary(ffile)
        # first record is int 3
        next(fb)
        n = fb.readbuf(1, 'i')[0]
        self.assertEqual(n, 3)

        # first record is float 1. 2. 3.
        next(fb)
        xref = (1., 2., 3.)
        x = fb.readbuf(n, 'd')
        np.testing.assert_allclose(x, xref)
        fb.close()

    def test_2(self):
        """Find and read label

          character*5 lab
          integer n
          lab = 'LABEL'
          n = 0
          open(1, file='fort.2', status='new', form='unformatted')
          write(1) n
          write(1) lab
          close(1)
          end
        """
        ffile = os.path.join(self.tdir, 'fort.2')
        fb = FortranBinary(ffile)
        rec  = fb.find(b'LABEL')

        self.assertEqual(rec.data, b'LABEL')
        fb.close()

    def test_2b(self):
        """Handle label not found

          character*5 lab
          integer n
          lab = 'LABEL'
          n = 0
          open(1, file='fort.2', status='new', form='unformatted')
          write(1) n
          write(1) lab
          close(1)
          end
        """
        ffile = os.path.join(self.tdir, 'fort.2')
        fb = FortranBinary(ffile)
        rec  = fb.find(b'NOLABEL')
        fb.close()

        self.assertEqual(rec, None)

    def test_3a(self):
        """Integer*8 dimensions

          integer*8, parameter :: nx = 3, ny=3
          double precision x(nx), y(ny)
          x = (/ 1.0D0, 2.0D0, 3.0D0 /)
          y = (/ 5.0D0, 6.0D0, 7.0D0 /)
          open(3, file='fort.3', status='new', form='unformatted')
          write(3) nx, ny
          write(3) x
          write(3) y
          close(3)
          end

        """
        ffile = os.path.join(self.tdir, 'fort.3')
        fb = FortranBinary(ffile)
        # first record is int 3, 3
        nx, ny = fb.next().read('q', 2)
        np.testing.assert_allclose((nx, ny), (3, 3))
        fb.close()

    def test_3b(self):
        """Read vecs

          integer, parameter :: nx = 3, ny=3
          double precision x(nx), y(ny)
          x = (/ 1.0D0, 2.0D0, 3.0D0 /)
          y = (/ 5.0D0, 6.0D0, 7.0D0 /)
          open(3, file='fort.3', status='new', form='unformatted')
          write(3) nx, ny
          write(3) x
          write(3) y
          close(3)
          end

        """
        ffile = os.path.join(self.tdir, 'fort.3')
        fb = FortranBinary(ffile)
        # first record is int 3
        fb.next()
        x=[]
        for rec in fb:
            x += list(fb.readbuf(3, 'd'))
        xref = (1., 2., 3.,  5., 6., 7.)
        np.testing.assert_allclose(x, xref)
        fb.close()

    def test_4(self):
        """Read string"""
        ffile = os.path.join(self.tdir, 'fort.4')
        fb = FortranBinary(ffile)
        rec = fb.find('ABC')
        self.assertIn(b'ABC', rec)

    def test_4b(self):
        """Read string"""
        ffile = os.path.join(self.tdir, 'fort.4')
        fb = FortranBinary(ffile)
        rec = fb.find(b'ABC')
        self.assertIn(b'ABC', rec)
        fb.close()

    def test_4c(self):
        """Read string"""
        ffile = os.path.join(self.tdir, 'fort.4')
        fb = FortranBinary(ffile)
        with self.assertRaises(ValueError):
            rec = fb.find(1.0)
        fb.close()


    def test_count_records_and_lengths(self):
        ffile = os.path.join(self.tdir, 'fort.3')
        fb = FortranBinary(ffile)
        self.assertTupleEqual(fb.record_byte_lengths(), (16, 24, 24))
        fb.close()

    def test_open_non_existing(self):
        ffile = os.path.join(self.tdir, 'nofile')
        with self.assertRaises(IOError):
            fb = FortranBinary(ffile)

    @mock_it('open')
    def test_open_new(self, mock_open):
        ffile = os.path.join(self.tdir, 'newfile')
        fb = FortranBinary(ffile, 'new')
        mock_open.assert_called_once_with(ffile, 'wb')


if __name__ == "__main__": #pragma: no cover
    unittest.main()
